from datetime import datetime
import re
import cv2
import numpy as np
from google.cloud import vision_v1
from idvpackage.constants import *
import pkg_resources
import base64
from googletrans import Translator
from rapidfuzz import fuzz

translator = Translator()

def create_final_result(dictionary):
    result = ""
    for key, value in dictionary.items():
        if isinstance(value, dict):
            sub_result = create_final_result(value)
            if sub_result == 'consider':
                return 'consider'
            elif sub_result == 'clear':
                result = 'clear'
        elif value in ['clear', 'consider', ""]:
            if value == 'consider':
                return 'consider'
            elif result != 'clear':
                result = value
    return result

def create_sub_result(document_report):
    sub_result = 'clear'

    digital_document = document_report["breakdown"]["image_integrity"]["breakdown"]["conclusive_document_quality"]["properties"].get("digital_document")
    corner_removed = document_report["breakdown"]["image_integrity"]["breakdown"]["conclusive_document_quality"]["properties"].get("corner_removed")
    watermarks_digital_text_overlay = document_report["breakdown"]["image_integrity"]["breakdown"]["conclusive_document_quality"]["properties"].get("watermarks_digital_text_overlay")
    obscured_security_features = document_report["breakdown"]["image_integrity"]["breakdown"]["conclusive_document_quality"]["properties"].get("obscured_security_features")
    screenshot = document_report["breakdown"]["visual_authenticity"]["breakdown"]["original_document_present"]["properties"].get("screenshot")
    document_on_printed_paper = document_report["breakdown"]["visual_authenticity"]["breakdown"]["original_document_present"]["properties"].get("document_on_printed_paper")
    photo_of_screen = document_report["breakdown"]["visual_authenticity"]["breakdown"]["original_document_present"]["properties"].get("photo_of_screen")
    scan = document_report["breakdown"]["visual_authenticity"]["breakdown"]["original_document_present"]["properties"].get("scan")

    consider_caution_count = sum(value == 'consider' for value in [digital_document, corner_removed, watermarks_digital_text_overlay, obscured_security_features, screenshot, document_on_printed_paper, photo_of_screen, scan])
    
    data_consistency = document_report["breakdown"]["data_consistency"]["result"]
    data_comparison = document_report["breakdown"]["data_comparison"]["result"]

    if data_consistency == 'consider' or data_comparison == 'consider':
        sub_result = 'caution'

    if consider_caution_count > 2:
        sub_result = 'suspected'

    return sub_result


def mark_clear(data, keys):
    if isinstance(data, dict):
        for key, value in list(data.items()):
            if key in keys:
                data[key] = mark_nested_result(value, 'clear')
            else:
                data[key] = mark_clear(value, keys)
        return data
    elif isinstance(data, list):
        return [mark_clear(item, keys) for item in data]
    else:
        return data


def mark_nested_result(data, value):
    if isinstance(data, dict):
        if 'result' in data:
            data['result'] = value
        for key, val in data.items():
            if key != 'result':
                data[key] = mark_nested_result(val, value)
        return data
    elif isinstance(data, list):
        return [mark_nested_result(item, value) for item in data]
    else:
        return data
    

def age_validation(dob, age_threshold=18):
    age_val = {
        "breakdown": {
            "minimum_accepted_age": {
            "properties": {},
            "result": ""
            }
        },
        "result": ""
        }
        
    dob = get_dates_to_generic_format(dob)
    try:
        dob_date = datetime.strptime(dob, "%d/%m/%Y")
    except:
        try:
            dob_date = datetime.strptime(dob, "%Y-%m-%d")
        except:
            dob_date = ''

    if dob_date:
        current_date = datetime.now()

        age = current_date.year - dob_date.year - ((current_date.month, current_date.day) < (dob_date.month, dob_date.day))

        if age>=age_threshold:
            age_val["breakdown"]["minimum_accepted_age"]["result"] = "clear"
            age_val["result"] = "clear"
        else:
            age_val["breakdown"]["minimum_accepted_age"]["result"] = "consider"
            age_val["result"] = "consider"
    else:
        age_val["breakdown"]["minimum_accepted_age"]["result"] = "consider"
        age_val["result"] = "consider"

    return age_val
    
def created_at():
    current_datetime = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')

    return current_datetime

def is_valid_and_not_expired(expiry_date, country):
    if country == 'LBN':    
        return 'clear'
        
    try:
        parsed_date = datetime.strptime(expiry_date, "%d/%m/%Y")
        current_date = datetime.now()
        
        if parsed_date < current_date:
            return 'consider'
        return 'clear'
    
    except:
        try:
            parsed_date = datetime.strptime(expiry_date, "%Y-%m-%d")
            current_date = datetime.now()
            
            if parsed_date < current_date:
                return 'consider'
            return 'clear'

        except:
            return 'consider'

def identify_document_type(text, country, origial_text=None):
    text = text.upper()
    emirates_id_pattern = r'\b(ILARE\w*|IDARE\w*|RESIDENT IDENTITY)\b'
    iqama_id_pattern = r'KINGDOM OF SAUDI ARABIA|RESIDENT IDENTITY|MINISTRY OF INTERIOR'
    iraq_id_pattern = r'Register|Signature|Family Number|The Directorate of Nationality|IDIRQ|The Republic of Iraq|Ministry of|National Card|Iraq'
    lbn_id_pattern = r'Marital status|Family|Lebanon Republic|Republic of Lebanon'
    qtr_id_pattern = r'State of Qatar|Residency Permit|Director General of the General Department|Directorate of Passports|Passport number'
    passport_pattern = r'\b(PASSPORT|PPT)\b'
    driver_license_pattern = r'\b(DRIVER|LICENSE|DL)\b'
    
    if re.search(emirates_id_pattern, text, re.IGNORECASE):
        return "EID"

    if country == 'IRQ' and (re.search(iraq_id_pattern, text, re.IGNORECASE) or re.search(iraq_id_pattern, origial_text, re.IGNORECASE)):
        return "INC"
    
    if country == 'LBN' and re.search(lbn_id_pattern, text, re.IGNORECASE):
        return "LBN"
    
    if country == 'QAT' and re.search(qtr_id_pattern, text, re.IGNORECASE):
        return "QAT"
        
    if re.search(passport_pattern, text, re.IGNORECASE):
        return "PASSPORT"

    if re.search(driver_license_pattern, text, re.IGNORECASE):
        return "DL"

    return "Unknown"

def identify_front_id(text):
    front_id_keywords = ['Resident Identity', 'United arab emirates', 'federal authority for identity', 'ID Number', 'Kingdom of saudi arabia', 'ministry of']
    pattern = '|'.join(map(re.escape, front_id_keywords))
    
    if re.search(pattern, text, re.IGNORECASE):
        return True
    else:
        return False

def identify_back_id(text):
    back_id_keywords = ['ILARE', 'IDARE', 'Signature']
    pattern = '|'.join(map(re.escape, back_id_keywords))
    
    if re.search(pattern, text, re.IGNORECASE):
        return True
    else:
        return False

def identify_front_id_iraq(text):
    front_id_keywords = ['The Republic of Iraq', 'Ministry of', 'National Card', 'Passports and Residence', 'Republic', 'Ministry', 'Iraq']
    pattern = '|'.join(map(re.escape, front_id_keywords))
    
    if re.search(pattern, text, re.IGNORECASE):
        return True
    else:
        return False

def identify_back_id_iraq(back_id_text, text):
    back_id_keywords = ['IDIRQ', 'Signature', 'Register', 'Family Number', 'The Directorate of Nationality', 'IDIR']
    pattern = '|'.join(map(re.escape, back_id_keywords))
    
    if re.search(pattern, text, re.IGNORECASE):
        return True
    else:
        if re.search(pattern, back_id_text, re.IGNORECASE):
            return True
        else:
            return False

def identify_front_id_lebanon(text):
    front_id_keywords = ['Lebanon Republic', 'Ministry of', 'Republic of Lebanon']
    pattern = '|'.join(map(re.escape, front_id_keywords))
    
    if re.search(pattern, text, re.IGNORECASE):
        return True
    else:
        return False

def identify_back_id_lebanon(text):
    back_id_keywords = ['Marital status', 'Family']
    pattern = '|'.join(map(re.escape, back_id_keywords))
    
    if re.search(pattern, text, re.IGNORECASE):
        return True
    else:
        return False

def identify_front_id_qatar(text):
    front_id_keywords = ['State of Qatar', 'Residency Permit']
    pattern = '|'.join(map(re.escape, front_id_keywords))
    
    if re.search(pattern, text, re.IGNORECASE):
        return True
    else:
        return False

def identify_back_id_qatar(text):
    back_id_keywords = ['Director General of the General Department', 'Directorate of Passports', 'Passport number']
    pattern = '|'.join(map(re.escape, back_id_keywords))
    
    if re.search(pattern, text, re.IGNORECASE):
        return True
    else:
        return False

def document_on_printed_paper(image, block_size=11, c_value=2, texture_threshold=0.4, contour_area_threshold=5000):
    image_gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    image_thresh = cv2.adaptiveThreshold(image_gray, 255, cv2.ADAPTIVE_THRESH_MEAN_C, cv2.THRESH_BINARY, block_size, c_value)
    image_contours, _ = cv2.findContours(image_thresh, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    
    # Filter out small noise contours
    image_paper_edges = [contour for contour in image_contours if cv2.contourArea(contour) > contour_area_threshold]

    # Calculate the percentage of paper texture in the image
    texture_ratio = len(image_paper_edges) / max(len(image_contours), 1)

    # Determine if the image contains a printed paper texture
    if texture_ratio >= texture_threshold:
        return 'consider'
    else:
        return 'clear'


def check_logo_existence_iraq(logo_content, id_card_content, threshold=3, good_mathces_filter=0.6):
    nparr_logo = np.frombuffer(logo_content, np.uint8)
    logo = cv2.imdecode(nparr_logo, cv2.IMREAD_GRAYSCALE)

    nparr_id_card = np.frombuffer(id_card_content, np.uint8)
    id_card = cv2.imdecode(nparr_id_card, cv2.IMREAD_COLOR)
    id_card_gray = cv2.cvtColor(id_card, cv2.COLOR_BGR2GRAY)

    sift = cv2.SIFT_create()

    kp1, des1 = sift.detectAndCompute(logo, None)
    kp2, des2 = sift.detectAndCompute(id_card_gray, None)

    FLANN_INDEX_KDTREE = 1
    index_params = dict(algorithm=FLANN_INDEX_KDTREE, trees=5)
    search_params = dict(checks=50)
    flann = cv2.FlannBasedMatcher(index_params, search_params)

    matches = flann.knnMatch(des1, des2, k=2)

    good_matches = []
    for m, n in matches:
        if m.distance < good_mathces_filter * n.distance:
            good_matches.append(m)

    # 40
    min_good_matches = threshold
    # min_good_matches = 3

    if len(good_matches) >= min_good_matches:
        return True
    else:
        return False


def detect_logo(client, input_image_content, country, compare_type=None, side=None):
    if country == 'UAE':
        reference_logo_path = pkg_resources.resource_filename('idvpackage', 'template_images/emirates_id_logo.jpeg')

        with open(reference_logo_path, 'rb') as logo_file:
            reference_image_content = logo_file.read()

        reference_image = vision_v1.types.Image(content=reference_image_content)
        input_image = vision_v1.types.Image(content=input_image_content)

        reference_response = client.logo_detection(image=reference_image)
        input_response = client.logo_detection(image=input_image)

        reference_logos = reference_response.logo_annotations
        input_logos = input_response.logo_annotations

        for reference_logo in reference_logos:
            for input_logo in input_logos:
                if reference_logo.description.lower() == input_logo.description.lower():
                    return 'clear'

        return 'consider'

    if country in ['IRQ', 'LBN', 'QAT'] and compare_type != 'template':
        if country == 'IRQ':
            logo_path = 'template_images/iraq_id_logo.png'
            good_mathces_filter = 0.6
        elif country == 'QAT':
            logo_path = 'template_images/qatar_id_logo.jpeg'
            good_mathces_filter = 0.7

        reference_logo_path = pkg_resources.resource_filename('idvpackage', logo_path)
        with open(reference_logo_path, 'rb') as logo_file:
            reference_image_content = logo_file.read()
        result = check_logo_existence_iraq(reference_image_content, input_image_content, 3, good_mathces_filter)

        if result:
            return 'clear'
        else:
            return 'consider'
    
    if country in ['IRQ', 'LBN', 'QAT'] and compare_type == 'template':
        if side == 'front':
            if country == 'QAT':
                threshold = 300
                front_template_path = 'template_images/qatar_front_standard_template.jpeg'
            elif country == 'IRQ':
                threshold = 40
                front_template_path = 'template_images/iraq_standard_template.png'
            else:  # LBN
                threshold = 60
                front_template_path = 'template_images/lbn_front_standard_template.png'
            reference_logo_path = pkg_resources.resource_filename('idvpackage', front_template_path)
        if side == 'back':
            if country == 'QAT':
                threshold = 235
                back_template_path = 'template_images/qatar_back_standard_template.jpeg'
            elif country == 'IRQ':
                threshold = 140
                back_template_path = 'template_images/iraq_back_standard_template.jpg'
            else:  # LBN
                threshold = 120
                back_template_path = 'template_images/lbn_back_standard_template.png'
            reference_logo_path = pkg_resources.resource_filename('idvpackage', back_template_path)

        with open(reference_logo_path, 'rb') as logo_file:
            reference_image_content = logo_file.read()

        result = check_logo_existence_iraq(reference_image_content, input_image_content, threshold)

        if result:
            return 'clear'
        else:
            return 'consider'


# def detect_logo(client, input_image_content, country):
#     if country == 'UAE':
#         reference_logo_path = pkg_resources.resource_filename('idvpackage', 'emirates_id_logo.jpeg')
    
        
#         with open(reference_logo_path, 'rb') as logo_file:
#             reference_image_content = logo_file.read()

#         reference_image = vision_v1.types.Image(content=reference_image_content)
#         input_image = vision_v1.types.Image(content=input_image_content)

#         reference_response = client.logo_detection(image=reference_image)
#         input_response = client.logo_detection(image=input_image)

#         reference_logos = reference_response.logo_annotations
#         input_logos = input_response.logo_annotations

#         for reference_logo in reference_logos:
#             for input_logo in input_logos:
#                 if reference_logo.description.lower() == input_logo.description.lower():
#                     return 'clear'

#         return 'consider'

#     if country == 'IRQ':
#         reference_logo_path = pkg_resources.resource_filename('idvpackage', 'iraq_id_logo.png')
#         result = check_logo_existence_iraq(reference_logo_path, input_image_content)

#         if result:
#             return 'clear'
#         else:
#             return 'consider'


def detect_logo_saudi(client, input_image):
    reference_logo_path = pkg_resources.resource_filename('idvpackage', 'template_images/sau_id_logo.png')
    
    with open(reference_logo_path, 'rb') as logo_file:
        reference_image_content = logo_file.read()

    reference_image = vision_v1.types.Image(content=reference_image_content)
    input_image = vision_v1.types.Image(content=input_image)

    reference_response = client.logo_detection(image=reference_image)
    input_response = client.logo_detection(image=input_image)

    reference_logos = reference_response.logo_annotations
    input_logos = input_response.logo_annotations

    for reference_logo in reference_logos:
        for input_logo in input_logos:
            if reference_logo.description.lower() == input_logo.description.lower():
                return 'clear'

    return 'consider'


def perform_feature_matching(image, template_image_path):
    template_image = cv2.imread(template_image_path, cv2.IMREAD_COLOR)

    sift = cv2.SIFT_create()

    kp1, des1 = sift.detectAndCompute(template_image, None)
    kp2, des2 = sift.detectAndCompute(image, None)

    if len(kp1) == 0 or len(kp2) == 0:
        return 0.0 
    
    bf = cv2.BFMatcher()
    matches = bf.knnMatch(des1, des2, k=2)

    good_matches = []
    for m, n in matches:
        if m.distance < 0.75 * n.distance:
            good_matches.append(m)

    similarity_score = len(good_matches) / len(kp1)

    return similarity_score


# part of detecting screenshot - checks for time values seen on phone
def extract_time_values(text_annotations):
    time_values = []
    for annotation in text_annotations:
        text = annotation.description
        time_matches = re.findall(r'\d{1,2}:\d{2}', text)
        time_values.extend(time_matches)
    return time_values


def detect_screenshot(client, image):
    icons = 0
    battery_value_and_time = 0

    similarity_threshold = 0.45

    for filename in pkg_resources.resource_listdir('idvpackage', 'icons'):
        if filename.endswith('.png'):
            icon_path = pkg_resources.resource_filename('idvpackage', f'icons/{filename}')
            image_data = np.frombuffer(base64.b64decode(image), dtype=np.uint8)

            front_similarity_score = perform_feature_matching(image_data, icon_path)
            if front_similarity_score >= similarity_threshold:
                icons+=1

    image_data = vision_v1.Image(content=base64.b64decode(image))

    image_response = client.text_detection(image=image_data)

    image_text_annotations = image_response.text_annotations

    image_time_values = extract_time_values(image_text_annotations)

    if image_time_values:
        battery_value_and_time+=1

    if icons or battery_value_and_time:
        return 'consider'
    else:
        return 'clear'


def detect_photo_on_screen(client, image):
    flag = 'clear'

    image_data = vision_v1.Image(content=base64.b64decode(image))
    
    image_response = client.label_detection(image=image_data)
    image_labels = image_response.label_annotations

    confidence_threshold = 0.58  

    # 'display device'

    keywords = ['mobile phone', 'mobile device', 'portable communications device', 'communication device', 'smartphone', 'cell phone', 'touchscreen', 'laptop', 'notebook', 'computer', 'screen', 'gadget']

    for label in image_labels:
        description = label.description.lower()
        confidence = label.score
        # print(f"Description: {description}")
        # print(f"Confidence: {confidence}")
        if confidence >= confidence_threshold:
            match = any(fuzz.ratio(description, keyword.lower()) >= 90 for keyword in keywords)
            if match:
                return 'consider'

    return flag

def fuzzy_match_fields(field1, field2, threshold=55):
    similarity = fuzz.partial_ratio(field1, field2)
    # print(f"similarity: {similarity}")
    return similarity >= threshold

def standardize_date(input_date):
        input_formats = [
            "%Y-%m-%d", "%m-%d-%Y", "%Y%m%d",
            "%Y/%m/%d", "%m/%d/%Y", "%d/%m/%Y",
            "%Y.%m.%d", "%d.%m.%Y", "%m.%d.%Y",
            "%Y %m %d", "%d %m %Y", "%m %d %Y",
        ]

        for format in input_formats:
            try:
                parsed_date = datetime.strptime(input_date, format)
                standardized_date = parsed_date.strftime("%d/%m/%Y")
                print(f"\n\n---------------Standardized date: { standardized_date}")
                return standardized_date
            except ValueError:
                pass

        return None

def compare_dates(date_str1, date_str2):
    date_format = "%d/%m/%Y"

    date1 = datetime.strptime(date_str1, date_format)
    date2 = datetime.strptime(date_str2, date_format)

    if date1 == date2:
        return True
    else:
        return False

def standardize_gender(gender):
    if gender == 'male' or gender == 'm':
        return 'Male'

    elif gender == 'female' or gender == 'f':
        return 'Female'

## for data comparison, accept data from dev and match with data extracted from id
def data_comparison_check(data, country):
    data_comparison = DATA_COMPARISON
    data_comparison = mark_clear(data_comparison, ['breakdown', 'result'])

    user_data = data.get('manual_input', '')

    if user_data and country not in ['IRQ', 'LBN', 'SDN', 'SYR', 'JOR', 'PSE']:
        if user_data.get('dob', '') and data.get('dob', ''):
            if compare_dates(get_dates_to_generic_format(data.get('dob', '')), get_dates_to_generic_format(user_data.get('dob', ''))):
                data_comparison['breakdown']['date_of_birth']['result'] = 'clear'
            else:
                data_comparison['breakdown']['date_of_birth']['result'] = 'consider'
        
        if user_data.get('gender', '') and data.get('gender', ''):    
            user_gender = standardize_gender(user_data.get('gender', '').lower())
            id_gender = standardize_gender(data.get('gender', '').lower())

            if user_gender == id_gender:
                data_comparison['breakdown']['gender']['result'] = 'clear'
            else:
                data_comparison['breakdown']['gender']['result'] = 'consider'

        # if not fuzzy_match_fields(data.get('gender', '').lower(),user_data.get('gender', '').lower()):
        #     data_comparison['breakdown']['gender']['result'] = 'consider'
    
        if country == 'IRQ':
            if data.get('doc_type') == 'national_identity_card':
                name_parts = data.get('name_en', '').split()
                if len(name_parts) >= 1:
                    first_name = name_parts[0]
                    last_name = name_parts[-1]
                else:
                    first_name, last_name = '',''

            elif data.get('doc_type') == 'passport':
                first_name = data.get('full_name', '').split()[0]
                last_name = data.get('last_name', '')

            else:
                name_parts = data.get('full_name', '').split()
                if len(name_parts) >= 1:
                    first_name = name_parts[0]
                    last_name = name_parts[-1]
                else:
                    first_name, last_name = '',''

        elif country == 'LBN':
            if data.get('doc_type') == 'passport':
                first_name = data.get('first_name', '')
                last_name = data.get('last_name', '')
            else:
                first_name = data.get('first_name_en', '')
                last_name = data.get('last_name_en', '')

        else:
            first_name = data.get('first_name', '')
            last_name = data.get('last_name', '')

        if user_data.get('first_name', ''):
            # print(f'\n\n--------------------------- Extracted first name: {first_name}')
            # print(f"--------------------------- User first name: {user_data.get('first_name', '')}")

            if fuzzy_match_fields(first_name.lower(), user_data.get('first_name', '').lower()):
                data_comparison['breakdown']['first_name']['result'] = 'clear'
            else:
                data_comparison['breakdown']['first_name']['result'] = 'consider'

        if user_data.get('last_name', ''):
            # print(f"\n--------------------------- Extracted last name: {last_name}")
            # print(f"--------------------------- User last name: {user_data.get('last_name', '')}")

            if fuzzy_match_fields(last_name.lower().replace('-', '').replace('_', ''), user_data.get('last_name', '').lower().replace('-', '').replace('_', '')):
                data_comparison['breakdown']['last_name']['result'] = 'clear'
            else:
                data_comparison['breakdown']['last_name']['result'] = 'consider'
        
        if country == 'QAT':
            data_comparison['breakdown']['date_of_birth']['result'] = 'clear'
            data_comparison['breakdown']['gender']['result'] = 'clear'
            data_comparison['breakdown']['first_name']['result'] = 'clear'
            data_comparison['breakdown']['last_name']['result'] = 'clear'

    ## set default to avoid any problems of cache
    data_comparison['result'] = 'clear'

    result = create_final_result(data_comparison)    
    data_comparison['result'] = result

    # print(f"\n\n---------------------------- DATA COMPARISON: {data_comparison}")

    return data_comparison

def data_consistency_check(data, front_id_text, back_id_text, country, back_img):
    from deep_translator import GoogleTranslator

    data_consistency = DATA_CONSISTENCY
    data_consistency = mark_clear(data_consistency, ['breakdown', 'result'])

    passport_data = data.get('passport')
    if passport_data:
        if not data.get('dob') == passport_data.get('passport_date_of_birth'):
            data_consistency['breakdown']['date_of_birth']['result'] = 'consider'
        
        if not fuzzy_match_fields(data.get('first_name').lower(),passport_data.get('passport_given_name').lower()):
            data_consistency['breakdown']['first_name']['result'] = 'consider'
        
        if not fuzzy_match_fields(data.get('gender').lower(),passport_data.get('passport_gender').lower()):
            data_consistency['breakdown']['gender']['result'] = 'consider'

        if not fuzzy_match_fields(data.get('last_name').lower(),passport_data.get('passport_surname').lower()):
            data_consistency['breakdown']['last_name']['result'] = 'consider'

    #### For data consistency compare data from different sources, like id and passport. 
    #### so the dob from id should match with dob extracted from passport

    if country == 'UAE':
        doc_type1 = identify_document_type(front_id_text, country)
        doc_type2 = identify_document_type(back_id_text, country)
        if doc_type1 == 'EID' or doc_type2=='EID':
            data_consistency['breakdown']['document_type']['result'] = 'clear'
        else:
            data_consistency['breakdown']['document_type']['result'] = 'consider'

    if country == 'SAU':
        doc_type1 = identify_front_id(front_id_text)
        if doc_type1:
            data_consistency['breakdown']['document_type']['result'] = 'clear'
        else:
            data_consistency['breakdown']['document_type']['result'] = 'consider'
    
    if country == 'IRQ':
        if data.get('doc_type') == 'passport':
            if 'passport' and ('iraq' or 'republic of iraq') in data.get('passport_data').lower():
                data_consistency['breakdown']['document_type']['result'] = 'clear'
            else:
                data_consistency['breakdown']['document_type']['result'] = 'clear'
        else:
            ## TODO: either completely remove this or make it efficient so we don't do the keyword checks everytime
            # try:
            #     translated_front_id_text = data['translated_front_id_text']
            #     translated_back_id_text = data['translated_back_id_text']
            # except:
            #     translated_front_id_text = GoogleTranslator('ar', 'en').translate(front_id_text)
            #     translated_back_id_text = GoogleTranslator('ar', 'en').translate(back_id_text)

            # doc_type1 = identify_document_type(translated_front_id_text, country, front_id_text)
            # doc_type2 = identify_document_type(translated_back_id_text, country, back_id_text)
            # if doc_type1 == 'INC' or doc_type2=='INC':
            #     data_consistency['breakdown']['document_type']['result'] = 'clear'
            # else:
            #     data_consistency['breakdown']['document_type']['result'] = 'consider'
            data_consistency['breakdown']['document_type']['result'] = 'clear'

        front_id_number = data.get('id_number_front', '')
        back_id_number = data.get('id_number', '')

        if front_id_number and back_id_number:
            if front_id_number != back_id_number:
                data_consistency['breakdown']['multiple_data_sources_present']['result'] = 'consider'

    if country == 'QAT':
        data_consistency['breakdown']['gender']['result'] = 'clear'
        data_consistency['breakdown']['last_name']['result'] = 'clear'
        data_consistency['breakdown']['first_name']['result'] = 'clear'
        data_consistency['breakdown']['nationality']['result'] = 'clear'
        data_consistency['breakdown']['date_of_birth']['result'] = 'clear'
        data_consistency['breakdown']['document_type']['result'] = 'clear'
        data_consistency['breakdown']['date_of_expiry']['result'] = 'clear'
        data_consistency['breakdown']['issuing_country']['result'] = 'clear'
        data_consistency['breakdown']['document_numbers']['result'] = 'clear'
        data_consistency['breakdown']['multiple_data_sources_present']['result'] = 'clear'

        id_number = data.get('id_number', '')
        serial_number = data.get('card_number', '')

        if id_number and serial_number:
            id_number_from_sr_no = serial_number[-11:]

            if id_number == id_number_from_sr_no:
                data_consistency['breakdown']['multiple_data_sources_present']['result'] = 'clear'
            else:
                print(f'Matching face from front and back')
                ## match faces from front and back
                from idvpackage.common import load_and_process_image_deepface, extract_face_and_compute_similarity
                front_face_locations, front_face_encodings = data.get('front_face_locations', []), data.get('front_face_encodings', [])
                try:
                    back_face_locations, back_face_encodings = load_and_process_image_deepface(back_img)
                    similarity = extract_face_and_compute_similarity(front_face_locations, front_face_encodings, back_face_locations, back_face_encodings)
                    print(f'\n\nSimilarity - front back: {similarity}\n')
                except Exception as e:
                    print(f'\n\nError: {e}\n')
                    similarity = 0.0
                    
                face_match = True if similarity >= 0.65 else False

                if face_match:
                    data_consistency['breakdown']['multiple_data_sources_present']['result'] = 'clear'
                else:
                    data_consistency['breakdown']['multiple_data_sources_present']['result'] = 'consider'

        else:
            data_consistency['breakdown']['multiple_data_sources_present']['result'] = 'consider'
            
        # translated_back_id_text = translator.translate(back_id_text, src='ar', dest='en').text 
        # doc_type1 = identify_document_type(front_id_text, country)
        # doc_type2 = identify_document_type(translated_back_id_text, country)
        # if doc_type1 == 'QAT' or doc_type2=='QAT':
        #     data_consistency['breakdown']['document_type']['result'] = 'clear'
        # else:
        #     data_consistency['breakdown']['document_type']['result'] = 'consider'

    if country in ['LBN', 'SDN', 'SYR', 'JOR', 'PSE']:
        data_consistency['breakdown']['document_type']['result'] = 'clear'

    ## set default to avoid any problems of cache
    data_consistency['result'] = 'clear'

    result = create_final_result(data_consistency)    
    data_consistency['result'] = result

    if country == 'QAT' and data_consistency['breakdown']['multiple_data_sources_present']['result'] != 'consider':
        data_consistency['result'] = 'clear'

    return data_consistency

def get_dates_to_generic_format(date):
    formats = ["%d/%m/%Y", "%Y/%m/%d", "%Y-%m-%d", "%d-%m-%Y"]
    for fmt in formats:
        try:
            return datetime.strptime(date, fmt).strftime("%d/%m/%Y")
        except ValueError:
            continue
    return date

def onfido_date_format(date):
    formats = ["%d/%m/%Y", "%Y/%m/%d"]
    for fmt in formats:
        try:
            return datetime.strptime(date, fmt).strftime("%Y-%m-%d")
        except ValueError:
            continue
    return date

def data_validation_check(data, country):
    data_validation = DATA_VALIDATION
    data_validation = mark_clear(data_validation, ['breakdown', 'result'])

    try:
        dob = data.get('dob')
        # print(f"DOB: {dob}")
        dob = get_dates_to_generic_format(dob)
        parsed_date = datetime.strptime(dob, "%d/%m/%Y")
        data_validation["breakdown"]['date_of_birth']["result"] = 'clear'
    except:
        try:
            dob = data.get('dob')
            parsed_date = datetime.strptime(dob, "%Y-%m-%d")
            data_validation["breakdown"]['date_of_birth']["result"] = 'clear'
        except:
            data_validation["breakdown"]['date_of_birth']["result"] = 'consider'

    try:
        doe = data.get('expiry_date')
        doe = get_dates_to_generic_format(doe)
        parsed_date = datetime.strptime(doe, "%d/%m/%Y")
        data_validation["breakdown"]['expiry_date']["result"] = 'clear'
    except:
        try:
            doe = data.get('expiry_date')
            parsed_date = datetime.strptime(doe, "%Y-%m-%d")
            data_validation["breakdown"]['expiry_date']["result"] = 'clear'
        except:
            data_validation["breakdown"]['expiry_date']["result"] = 'consider'

    gender = data.get('gender', '')
    
    if gender.isalpha() and (gender.lower()=='male' or gender.lower()=='female' or gender.lower()=='m' or gender.lower()=='f'):
        data_validation["breakdown"]['gender']["result"] = 'clear'
    else:
        data_validation["breakdown"]['gender']["result"] = 'consider'
    
    if country == 'LBN' and data.get('doc_type') == 'national_identity_card':
        data_validation["breakdown"]['expiry_date']["result"] = 'clear'
        data_validation['breakdown']['valid_nationality']["result"] = 'clear'
        data_validation['breakdown']['document_expiration']["result"] = 'clear'
    else:
        data_validation['breakdown']['valid_nationality']["result"] = data.get('valid_nationality')

    if country in ['SDN', 'SYR', 'JOR', 'PSE']:
        data_validation['breakdown']['valid_nationality']["result"] = data.get('valid_nationality')
        data_validation["breakdown"]['document_numbers']["result"] = 'clear'

    if country == 'UAE' or country == 'IRQ':
        if data.get('doc_type') == 'national_identity_card':
            doc_no = data.get('card_number', '')
            if len(doc_no)==9:
                data_validation["breakdown"]['document_numbers']["result"] = 'clear'
            else:
                data_validation["breakdown"]['document_numbers']["result"] = 'consider'
        else:
            data_validation["breakdown"]['document_numbers']["result"] = 'clear'

    doe = data.get('expiry_date', '')
    expiry_result = is_valid_and_not_expired(get_dates_to_generic_format(doe), country)
    data_validation['breakdown']['document_expiration']["result"] = expiry_result

    mrz = data.get('mrz', '')
    mrz1 = data.get('mrz1', '')
    mrz2 = data.get('mrz2', '')
    mrz3 = data.get('mrz3', '')

    if data.get('doc_type') == 'passport':
        if mrz and mrz1 and mrz2:
            data_validation["breakdown"]['mrz']["result"] = 'clear'
        else:
            data_validation["breakdown"]['mrz']["result"] = 'consider'
    else:
        if len(mrz) == 1 and mrz1 and mrz2 and mrz3:
            data_validation["breakdown"]['mrz']["result"] = 'clear'
        else:
            data_validation["breakdown"]['mrz']["result"] = 'consider'

    if country == 'LBN' and data.get('doc_type') == 'national_identity_card':
        doc_no = data.get('id_number', '')
        if len(doc_no)>=10 and len(doc_no)<13:
            data_validation["breakdown"]['document_numbers']["result"] = 'clear'
        else:
            data_validation["breakdown"]['document_numbers']["result"] = 'consider'
        
        data_validation["breakdown"]['mrz']["result"] = 'clear'

    if country == 'QAT':
        data_validation["breakdown"]['document_numbers']["result"] = 'clear'
        data_validation["breakdown"]['gender']["result"] = 'clear'
        data_validation["breakdown"]['mrz']["result"] = 'clear'
        data_validation["breakdown"]['document_expiration']["result"] = 'clear'
        data_validation['breakdown']['valid_nationality']["result"] = 'clear'
        data_validation["breakdown"]['date_of_birth']["result"] = 'clear'
        data_validation["breakdown"]['expiry_date']["result"] = 'clear'

    if country == 'IRQ' and data.get('doc_type') == 'national_identity_card' and not data.get('card_number'):
        doc_no = data.get('card_number_front', '')
        if len(doc_no)==9:
            data_validation["breakdown"]['document_numbers']["result"] = 'clear'
        else:
            data_validation["breakdown"]['document_numbers']["result"] = 'consider'

    # print(f"\n\n------------------------- DATA VAL: {data_validation}")
    
    ## set default to avoid any problems of cache
    data_validation['result'] = 'clear'

    result = create_final_result(data_validation)
    data_validation['result'] = result

    # if data_validation["breakdown"]['date_of_birth']["result"]=='clear' and data_validation["breakdown"]['expiry_date']["result"]=='clear' and data_validation["breakdown"]['gender']["result"]=='clear' and data_validation["breakdown"]['mrz']["result"]=='clear':
    #     data_validation['result'] = 'clear'

    return data_validation

## pending
def image_integrity_check(data, front_id_text, back_id_text, coloured, blurred, glare, missing_fields, country):
    from deep_translator import GoogleTranslator
    image_integrity = IMAGE_INTEGRITY
    image_integrity = mark_clear(image_integrity, ['breakdown', 'result'])

    image_integrity['breakdown']['colour_picture']['result'] = coloured
    image_integrity['breakdown']['conclusive_document_quality']['properties']['corner_removed'] = is_valid_and_not_expired(get_dates_to_generic_format(data.get('expiry_date', '')), country)
    image_integrity['breakdown']['conclusive_document_quality']['properties']['abnormal_document_features'] = blurred
    image_integrity['breakdown']['image_quality']['properties']['blurred_photo'] = blurred
    image_integrity['breakdown']['image_quality']['properties']['covered_photo'] = missing_fields
    image_integrity['breakdown']['image_quality']['properties']['cut_off_document'] = missing_fields
    image_integrity['breakdown']['image_quality']['properties']['glare_on_photo'] = glare
    image_integrity['breakdown']['image_quality']['properties']['other_photo_issue'] =  missing_fields
    
    ## set default to avoid any problems of cache
    image_integrity['breakdown']['image_quality']['result'] = 'clear'

    image_quality_result = create_final_result(image_integrity['breakdown']['image_quality'])
    image_integrity['breakdown']['image_quality']['result'] = image_quality_result

    f_result = 'clear'
    if data.get('doc_type') == 'national_identity_card':
        f_result = identify_front_id(front_id_text)
    
    front_doc_on_pp = data.get('front_doc_on_pp')

    if country == 'UAE':
        b_result = identify_back_id(back_id_text)

        if back_id_text and b_result:
            image_integrity['breakdown']['conclusive_document_quality']['properties']['missing_back'] = 'clear'
        else:
            image_integrity['breakdown']['conclusive_document_quality']['properties']['missing_back'] = 'consider'

        if f_result and b_result:
            image_integrity['breakdown']['supported_document']['result'] = 'clear'
        else:
            image_integrity['breakdown']['supported_document']['result'] = 'consider'
    
    if country == 'SAU':
        if f_result:
            image_integrity['breakdown']['supported_document']['result'] = 'clear'
        else:
            image_integrity['breakdown']['supported_document']['result'] = 'consider'
        
        image_integrity['breakdown']['conclusive_document_quality']['properties']['digital_document'] = front_doc_on_pp

    if country == 'IRQ' and data.get('doc_type') == 'national_identity_card':
        image_integrity['breakdown']['conclusive_document_quality']['properties']['missing_back'] = 'clear'
        image_integrity['breakdown']['supported_document']['result'] = 'clear'
        
        ## TODO: either completely remove this or make it efficient so we don't do the keyword checks everytime
        # try:
        #     translated_back_id_text = data['translated_back_id_text']
        # except:
        #     translated_back_id_text = GoogleTranslator('ar', 'en').translate(back_id_text)

        # b_result = identify_back_id_iraq(back_id_text, translated_back_id_text)

        # if back_id_text and b_result:
        #     image_integrity['breakdown']['conclusive_document_quality']['properties']['missing_back'] = 'clear'
        # else:
        #     image_integrity['breakdown']['conclusive_document_quality']['properties']['missing_back'] = 'consider'
        
        # try:
        #     translated_front_id_text = translator.translate(front_id_text, src='ar', dest='en').text 
        # except:
        #     translated_front_id_text = GoogleTranslator('ar', 'en').translate(front_id_text)

        # f_result = identify_front_id_iraq(translated_front_id_text)

        # if f_result and b_result:
        #     image_integrity['breakdown']['supported_document']['result'] = 'clear'
        # else:
        #     image_integrity['breakdown']['supported_document']['result'] = 'consider'

    elif country == 'IRQ' and data.get('doc_type') == 'passport':
        image_integrity['breakdown']['conclusive_document_quality']['properties']['missing_back'] = 'clear'
        if 'passport' and ('iraq' or 'republic of iraq') in data.get('passport_data').lower():
            image_integrity['breakdown']['supported_document']['result'] = 'clear'

    if country == 'LBN' and data.get('doc_type') == 'national_identity_card':
        try:
            translated_back_id_text = data['translated_back_id_text']
        except:
            translated_back_id_text = GoogleTranslator('ar', 'en').translate(back_id_text)

        b_result = identify_back_id_lebanon(translated_back_id_text)

        if back_id_text and b_result:
            image_integrity['breakdown']['conclusive_document_quality']['properties']['missing_back'] = 'clear'
        else:
            image_integrity['breakdown']['conclusive_document_quality']['properties']['missing_back'] = 'consider'
        
        try:
            translated_front_id_text = data['translated_front_id_text']
        except:
            translated_front_id_text = GoogleTranslator('ar', 'en').translate(front_id_text)
            
        f_result = identify_front_id_lebanon(translated_front_id_text)

        if f_result and b_result:
            image_integrity['breakdown']['supported_document']['result'] = 'clear'
        else:
            image_integrity['breakdown']['supported_document']['result'] = 'consider'

    elif country == 'LBN' and data.get('doc_type') == 'passport':
        image_integrity['breakdown']['conclusive_document_quality']['properties']['missing_back'] = 'clear'
        if 'passport' and ('lebanon' or 'republic of lebanon') in data.get('passport_data').lower():
            image_integrity['breakdown']['supported_document']['result'] = 'clear'

    image_integrity['breakdown']['conclusive_document_quality']['properties']['watermarks_digital_text_overlay'] = data.get('front_logo_result')
    # image_integrity['breakdown']['image_quality']['properties']['blurred_photo'] = blurred
    # image_integrity['breakdown']['image_quality']['properties']['glare_on_photo'] = glare

    if country == 'QAT':
        image_integrity['breakdown']['image_quality']['properties']['blurred_photo'] = 'clear'
        image_integrity['breakdown']['image_quality']['properties']['covered_photo'] = 'clear'
        image_integrity['breakdown']['image_quality']['properties']['cut_off_document'] = 'clear'
        image_integrity['breakdown']['image_quality']['properties']['glare_on_photo'] = 'clear'
        image_integrity['breakdown']['image_quality']['properties']['other_photo_issue'] =  'clear'

        image_integrity['breakdown']['colour_picture']['result'] = 'clear'

        image_integrity['breakdown']['supported_document']['result'] = 'clear'

        image_integrity['breakdown']['conclusive_document_quality']['properties']['watermarks_digital_text_overlay'] = 'clear'
        image_integrity['breakdown']['conclusive_document_quality']['properties']['missing_back'] = 'clear'
        image_integrity['breakdown']['conclusive_document_quality']['properties']['corner_removed'] = 'clear'
        image_integrity['breakdown']['conclusive_document_quality']['properties']['abnormal_document_features'] = 'clear'
        image_integrity['breakdown']['conclusive_document_quality']['properties']['digital_document'] = 'clear'
        image_integrity['breakdown']['conclusive_document_quality']['properties']['punctured_document'] = 'clear'
        image_integrity['breakdown']['conclusive_document_quality']['properties']['obscured_data_points'] = 'clear'
        image_integrity['breakdown']['conclusive_document_quality']['properties']['obscured_security_features'] = 'clear'

    image_quality_result = create_final_result(image_integrity['breakdown']['image_quality'])
    conclusive_document_quality_result = create_final_result(image_integrity['breakdown']['conclusive_document_quality'])

    ## set default to avoid any problems of cache
    image_integrity['breakdown']['conclusive_document_quality']["result"] = 'clear'

    image_integrity['breakdown']['conclusive_document_quality']["result"] = conclusive_document_quality_result

    colour_picture_result = image_integrity['breakdown']['colour_picture']['result']
    supported_documents_result = image_integrity['breakdown']['supported_document']['result']

    back_doc_on_pp = data.get('doc_on_pp')
    if front_doc_on_pp == 'consider' or back_doc_on_pp == 'consider':
        image_integrity['breakdown']['conclusive_document_quality']['properties']['digital_document'] = 'consider'

    if country == 'QAT' or country == 'IRQ' or country == 'LBN':
        image_integrity['breakdown']['conclusive_document_quality']['properties']['digital_document'] = 'clear'

        # translated_back_id_text = translator.translate(back_id_text, src='ar', dest='en').text 
        # b_result = identify_back_id_qatar(translated_back_id_text)

        # if back_id_text and b_result:
            # image_integrity['breakdown']['conclusive_document_quality']['properties']['missing_back'] = 'clear'
        # else:
        #     image_integrity['breakdown']['conclusive_document_quality']['properties']['missing_back'] = 'consider'
        
        # f_result = identify_front_id_qatar(front_id_text)

        # if f_result and b_result:
        #     image_integrity['breakdown']['supported_document']['result'] = 'clear'
        # else:
        #     image_integrity['breakdown']['supported_document']['result'] = 'consider'


    if country in ['SDN', 'SYR', 'JOR', 'PSE']:
        image_integrity['breakdown']['supported_document']['result'] = 'clear'
        image_integrity['breakdown']['conclusive_document_quality']['properties']['missing_back'] = 'clear'
        image_integrity['breakdown']['image_quality']['properties']['cut_off_document'] = 'clear'
        image_integrity['breakdown']['image_quality']['properties']['other_photo_issue'] =  'clear'
        image_integrity['breakdown']['image_quality']['properties']['blurred_photo'] = blurred
        image_integrity['breakdown']['conclusive_document_quality']['properties']['digital_document'] = 'clear'

    ## set default to avoid any problems of cache
    image_integrity['result'] = 'clear'

    if image_quality_result == 'consider' or conclusive_document_quality_result == 'consider' or supported_documents_result == 'consider':
        image_integrity['result'] = 'consider'

    return image_integrity

def visual_authenticity_check(data, front_id_text, back_id_text, selfie, facial_similarity, face_match_threshold, country):
    from deep_translator import GoogleTranslator
    visual_authenticity = VISUAL_AUTHENTICITY
    visual_authenticity = mark_clear(visual_authenticity, ['breakdown', 'result'])

    if np.any(selfie):
        ## if facial similarity is matching the threshold or even if facial similarity comes above 40 then we say face was detected - approach can be changed
        if facial_similarity >= face_match_threshold:
            visual_authenticity['breakdown']['face_detection'] = 'clear'
            visual_authenticity['breakdown']['security_features'] = 'clear'
        else:
            visual_authenticity['breakdown']['face_detection'] = 'consider'
            visual_authenticity['breakdown']['security_features'] = 'consider'
    else:
        visual_authenticity['breakdown']['face_detection'] = ''
        visual_authenticity['breakdown']['security_features'] = ''
    
    if data.get('doc_type') == 'national_identity_card':
        doc_type1 = identify_document_type(front_id_text, country, front_id_text)

    front_doc_on_pp = data.get('front_doc_on_pp')
    front_screenshot = data.get('front_screenshot_result')
    front_photo_on_screen_result = data.get('front_photo_on_screen_result')

    if country != 'SAU':
        back_screenshot = data.get('screenshot_result')
        if front_screenshot == 'consider' or back_screenshot == 'consider':
            visual_authenticity['breakdown']['original_document_present']['properties']['screenshot'] = 'consider'

        back_photo_on_screen_result = data.get('photo_on_screen_result')
        if front_photo_on_screen_result == 'consider' or back_photo_on_screen_result == 'consider':
            visual_authenticity['breakdown']['original_document_present']['properties']['photo_of_screen'] = 'consider'

    if country == 'UAE':
        doc_type2 = identify_document_type(back_id_text, country)
        if doc_type1 == 'EID' and doc_type2 == 'EID':
            visual_authenticity['breakdown']['original_document_present']['properties']['scan'] = 'clear'
        else:
            visual_authenticity['breakdown']['original_document_present']['properties']['scan'] = 'consider'

        visual_authenticity['breakdown']['face_detection'] = 'clear'
        visual_authenticity['breakdown']['security_features'] = 'clear'

    if country == 'SAU':
        if identify_front_id(front_id_text):
            visual_authenticity['breakdown']['original_document_present']['properties']['scan'] = 'clear'
        else:
            visual_authenticity['breakdown']['original_document_present']['properties']['scan'] = 'consider'
    
        visual_authenticity['breakdown']['original_document_present']['properties']['document_on_printed_paper'] = front_doc_on_pp
        visual_authenticity['breakdown']['original_document_present']['properties']['screenshot'] = front_screenshot
        visual_authenticity['breakdown']['original_document_present']['properties']['photo_of_screen'] = front_photo_on_screen_result

    if country == 'IRQ':
        if data.get('doc_type') == 'passport':
            if 'passport' and ('iraq' or 'republic of iraq') in data.get('passport_data').lower():
                visual_authenticity['breakdown']['original_document_present']['properties']['scan'] = 'clear'
            else:
                visual_authenticity['breakdown']['original_document_present']['properties']['scan'] = 'consider'
        else:
            try:
                translated_front_id_text = data['translated_front_id_text']
                translated_back_id_text = data['translated_back_id_text']
            except:
                translated_front_id_text = GoogleTranslator('ar', 'en').translate(front_id_text)
                translated_back_id_text = GoogleTranslator('ar', 'en').translate(back_id_text)

            doc_type1 = identify_document_type(translated_front_id_text, country, front_id_text)
            doc_type2 = identify_document_type(translated_back_id_text, country, back_id_text)
            if doc_type1 == 'INC' and doc_type2 == 'INC':
                visual_authenticity['breakdown']['original_document_present']['properties']['scan'] = 'clear'
            else:
                visual_authenticity['breakdown']['original_document_present']['properties']['scan'] = 'clear'

            visual_authenticity['breakdown']['template']['result'] = data.get('front_template_result')
            visual_authenticity['breakdown']['digital_tampering']['result'] = data.get('tampering_result')

    if country == 'LBN':
        visual_authenticity['breakdown']['original_document_present']['properties']['scan'] = 'clear'
        visual_authenticity['breakdown']['original_document_present']['properties']['screenshot'] = 'clear'
        visual_authenticity['breakdown']['digital_tampering']['result'] = 'clear'
        visual_authenticity['breakdown']['template']['result'] = 'clear'

    if country == 'QAT':
        visual_authenticity['breakdown']['original_document_present']['properties']['scan'] = 'clear'
        visual_authenticity['breakdown']['original_document_present']['properties']['screenshot'] = 'clear'
        visual_authenticity['breakdown']['original_document_present']['properties']['photo_of_screen'] = 'clear'
        visual_authenticity['breakdown']['original_document_present']['properties']['document_on_printed_paper'] = 'clear'
        visual_authenticity['breakdown']['digital_tampering']['result'] = 'clear'
        visual_authenticity['breakdown']['template']['result'] = 'clear'

        # translated_back_id_text = translator.translate(back_id_text, src='ar', dest='en').text 
        
        # doc_type1 = identify_document_type(front_id_text, country)
        # doc_type2 = identify_document_type(translated_back_id_text, country)
        # if doc_type1 == 'QAT' and doc_type2 == 'QAT':
            # visual_authenticity['breakdown']['original_document_present']['properties']['scan'] = 'clear'
        # else:
            # visual_authenticity['breakdown']['original_document_present']['properties']['scan'] = 'consider'

    if country in ['SDN', 'SYR', 'JOR', 'PSE']:
        visual_authenticity['breakdown']['original_document_present']['properties']['scan'] = 'clear'
        visual_authenticity['breakdown']['original_document_present']['properties']['screenshot'] = 'clear'
        visual_authenticity['breakdown']['digital_tampering']['result'] = 'clear'
        visual_authenticity['breakdown']['template']['result'] = 'clear'

    original_document_present_result  = create_final_result(visual_authenticity['breakdown']['original_document_present'])
    visual_authenticity['breakdown']['original_document_present']['result'] = original_document_present_result

    ## set default to avoid any problems of cache
    visual_authenticity['result'] = 'clear'

    final_result = create_final_result(visual_authenticity)
    visual_authenticity['result'] = final_result

    # print(f"\n\n------------------------- VISUAL AUTHENTICITY DOC: {visual_authenticity}")

    back_doc_on_pp = data.get('doc_on_pp')
    if front_doc_on_pp == 'consider' or back_doc_on_pp == 'consider':
        visual_authenticity['breakdown']['original_document_present']['properties']['document_on_printed_paper'] = 'consider'

    if country == 'QAT':
        visual_authenticity['breakdown']['original_document_present']['result'] = 'clear'
        visual_authenticity['breakdown']['original_document_present']['properties']['document_on_printed_paper'] = 'clear'
        visual_authenticity['result'] = 'clear'

    return visual_authenticity

def main_details(data, country):
    main_properties = {}
    
    main_properties['document_numbers']= []

    try:
        dob = get_dates_to_generic_format(data.get('dob'))
        doe = get_dates_to_generic_format(data.get('expiry_date', ''))
        main_properties['date_of_birth'] = onfido_date_format(dob)
        main_properties['date_of_expiry'] = onfido_date_format(doe)

        if data.get('card_number'):
            card_data_t = {
            "type": "type",
            "value": "document_number"
            }

            card_data_v = {
                "type": "value",
                "value": data['card_number']
            }

            main_properties['document_numbers'].append(card_data_t)
            main_properties['document_numbers'].append(card_data_v)

        if data.get('id_number'):
            id_data_t = {
                        "type": "type",
                        "value": "personal_number"
                    }
            id_data_v = {
                        "type": "value",
                        "value": data['id_number']
                    }
                
            main_properties['document_numbers'].append(id_data_t) 
            main_properties['document_numbers'].append(id_data_v) 

        else:
            id_data_t = {
                        "type": "type",
                        "value": "personal_number"
                    }
            id_data_v = {
                        "type": "value",
                        "value": ""
                    }
                
            main_properties['document_numbers'].append(id_data_t) 
            main_properties['document_numbers'].append(id_data_v)
        
        # main_properties['document_type'] = 'national_identity_card'
        # main_properties['name'] = data.get('name')
        # main_properties['first_name'] = data.get('first_name', '')
        # main_properties['gender'] = data.get('gender', '')
        # main_properties['issuing_country'] = data.get('issuing_place', '')
        # main_properties['last_name'] = data.get('last_name', '')
        # main_properties['mrz_line1'] = data.get('mrz1', '')
        # main_properties['mrz_line2'] = data.get('mrz2', '')
        # main_properties['mrz_line3'] = data.get('mrz3', '')
        # main_properties['nationality'] = data.get('nationality')

        keys = ['name', 'first_name', 'gender', 'issuing_place', 'last_name', 'nationality', 'issuing_country', 'occupation', 'employer', 'family_sponsor']

        main_properties['document_type'] = data.get('doc_type', '')
        main_properties['mrz_line1'] = data.get('mrz1', '')
        main_properties['mrz_line2'] = data.get('mrz2', '')
        main_properties['mrz_line3'] = data.get('mrz3', '')

        for key in keys:
            main_properties[key] = data.get(key, '')
        
        if country in ['IRQ', 'LBN', 'QAT', 'SDN', 'SYR', 'JOR', 'PSE']:
            if country == 'IRQ':
                main_prop_data = {}
                if data.get('doc_type') == 'passport':
                    keys_to_remove = ['issuing_place', 'occupation', 'employer', 'mrz_line3', 'family_sponsor', 'issuing_authority', 'issuing_authority_en', 'place_of_birth', 'place_of_birth_en', 'gender_ar', 'name_en', 'name', 'first_name']
                    keys = ['full_name', 'id_number', 'mrz', 'issue_date', 'mother_first_name', 'mother_first_name_en', 'mother_last_name', 'mother_last_name_en']
                else:
                    keys_to_remove = ['occupation', 'employer', 'family_sponsor', 'issuing_place']
                    keys = ['name_en', 'name', 'first_name', 'father_name', 'third_name', 'mother_first_name', 'mother_last_name', 'last_name', 'first_name_en', 'father_name_en', 'last_name_en', 'third_name_en', 'mother_first_name_en', 'mother_last_name_en', 'issuing_authority', 'nationality', 'gender', 'issuing_country', 'gender_ar', 'mrz', 'place_of_birth', 'issuing_authority_en', 'place_of_birth_en', "issue_date", "id_number", "card_number", 'family_number', 'family_number_en']

            if country == 'LBN':
                main_prop_data = {}
                keys_to_remove = ['name', 'issuing_place', 'occupation', 'employer', 'mrz_line3', 'family_sponsor']
                if data.get('doc_type') == 'passport':
                    keys = ['first_name', 'father_name', 'mother_name', 'last_name', 'id_number', 'issue_date', 'mrz', 'issuing_country', 'registry_place_and_number', 'place_of_birth', 'nationality']
                else:
                    # keys = ['father_name', 'last_name', 'name_en', 'first_name_en', 'fathers_name_en', 'last_name_en', 'gender_ar', 'place_of_birth', 'issue_date', 'issue_date_ar', 'card_number_ar', 'id_number_ar']
                    keys = ['name', 'name_en', 'first_name', 'father_name', 'last_name', 'first_name_en', 'father_name_en', 'last_name_en', 'mother_name_en', 'gender_ar', 'place_of_birth', 'place_of_birth_en', 'issue_date', 'issue_date_ar', 'card_number_ar', 'id_number_ar', 'issuing_country', 'nationality']
                    keys_to_remove = ['issuing_place', 'mrz_line1', 'mrz_line2', 'mrz_line3', 'family_sponsor', 'nationality', 'occupation', 'employer']

            if country == 'QAT':
                main_prop_data = {}
                keys_to_remove = ['gender', 'issuing_place', 'mrz_line1', 'mrz_line2', 'mrz_line3', 'name_en', 'family_sponsor']
                keys = ['passport_number', 'name_ar', 'passport_expiry', 'occupation', 'employer', 'employer_en', 'occupation_en']
            
            if country == 'SDN':
                main_prop_data = {}
                keys_to_remove = ['name', 'issuing_place', 'employer','mrz_line3', 'family_sponsor', 'first_name', 'last_name', 'father_name', 'mother_name']
                if data.get('doc_type') == 'passport':
                    keys = ['full_name_generic', 'name_ar', 'first_name', 'last_name', 'middle_name', 'first_name_ar', 'last_name_ar', 'middle_name_ar', 'gender', 'place_of_birth', 'place_of_issue', 'passport_number_mrz', 'national_number', 'issue_date', 'mrz', 'nationality', 'issuing_country']
                else:
                    # keys = ['father_name', 'last_name', 'name_en', 'first_name_en', 'fathers_name_en', 'last_name_en', 'gender_ar', 'place_of_birth', 'issue_date', 'issue_date_ar', 'card_number_ar', 'id_number_ar']
                    keys = ['name', 'name_ar', 'nationality', 'gender', 'place_of_birth', 'place_of_birth_en', 'id_number', 'first_name_ar','last_name_ar','middle_name_ar','occupation', 'occupation_en', 'occupation_ar','issue_date', 'first_name', 'middle_name', 'last_name', 'issuing_country']
                    keys_to_remove = ['issuing_place', 'family_sponsor', 'employer', 'first_name', 'last_name', 'father_name', 'mother_name', 'place_of_issue']


            for key in keys:
                main_prop_data[key] =  data.get(key, '')

            if keys_to_remove:
                for key in keys_to_remove:
                    if key in main_properties:
                        del main_properties[key]

            main_properties.update(main_prop_data)

            if main_properties.get('issue_date', ''):
                main_properties['issue_date'] = onfido_date_format(get_dates_to_generic_format(main_properties['issue_date']))

            if not data.get('card_number') and data.get('card_number_front'):
                card_data_t = {
                "type": "type",
                "value": "document_number"
                }

                card_data_v = {
                    "type": "value",
                    "value": data['card_number_front']
                }

                main_properties['document_numbers'].append(card_data_t)
                main_properties['document_numbers'].append(card_data_v)

            if not data.get('id_number') and data.get('id_number_front'):
                id_data_t = {
                            "type": "type",
                            "value": "personal_number"
                        }
                id_data_v = {
                            "type": "value",
                            "value": data['id_number_front']
                        }
                    
                main_properties['document_numbers'].append(id_data_t) 
                main_properties['document_numbers'].append(id_data_v) 

    except:
        main_properties

    return main_properties    


def form_final_data_document_report(data, front_id_text, back_id_text, country, coloured, selfie, facial_similarity, blurred, glare, missing_fields, face_match_threshold, back_img):
    try:
        document_report = {
            ## pending - to be filled by dev
            "_id": "",
            "breakdown": {
                "age_validation": age_validation(data.get('dob')),
                "compromised_document": {
                    "result": "clear"
                    },
                "data_comparison": data_comparison_check(data, country),
                "data_consistency": data_consistency_check(data, front_id_text, back_id_text, country, back_img),
                "data_validation": data_validation_check(data, country),
                "image_integrity": image_integrity_check(data, front_id_text, back_id_text, coloured, blurred, glare, missing_fields, country),
                "issuing_authority": {
                "breakdown": {
                    "nfc_active_authentication": {
                    "properties": {}
                    },
                    "nfc_passive_authentication": {
                    "properties": {}
                    }
                }
                },
                "police_record": {},
                "visual_authenticity": visual_authenticity_check(data, front_id_text, back_id_text, selfie, facial_similarity, face_match_threshold, country)
            },
            ## pending - to be filled by dev
            "check_id": "", 
            "created_at": created_at(),
            "documents": [
                {
                ## pending - id value in table stored in db for front id - to be filled by dev
                "id": ""
                },
                {
                ## pending - id value in table stored in db for front id - to be filled by dev
                "id": ""
                }
            ],
            # to be filled by dev
            "href": "",
            "name": "document",
            "properties": main_details(data, country),
            "result": "",
            "status": "complete",
            "sub_result": ""
        }
        
        final_result = create_final_result(document_report)
        document_report['result'] = final_result

        sub_result = create_sub_result(document_report)
        document_report['sub_result'] = sub_result

        if country == 'UAE':
            keys_to_mark = ['original_document_present','conclusive_document_quality','image_quality','data_consistency', 'data_comparison', 'image_integrity', 'data_validation', 'visual_authenticity']
            document_report = mark_clear(document_report, keys_to_mark)
            document_report['result'] = 'clear'
            document_report['sub_result'] = 'clear'

        if country == 'QAT' and(document_report['breakdown']['visual_authenticity']['result'] == 'clear' and document_report['breakdown']['visual_authenticity']['breakdown']['face_detection'] == 'clear' and document_report['breakdown']['data_consistency']['breakdown']['multiple_data_sources_present']['result'] == 'clear'):
            keys_to_mark = ['data_consistency', 'data_comparison', 'image_integrity', 'data_validation', 'visual_authenticity']
            document_report = mark_clear(document_report, keys_to_mark)
            document_report['result'] = 'clear'

        return document_report
    
    except Exception as e:
        print(e)
        return {}

def form_final_facial_similarity_report(data, selfie, facial_similarity, liveness_result, face_match_threshold, country):
    try:
        facial_report = FACIAL_REPORT
        facial_report = mark_clear(facial_report, ['breakdown'])
        
        facial_report['created_at'] = created_at()
        
        facial_report['breakdown']['face_comparison']['breakdown']['face_match']['properties']['score'] = facial_similarity

        if not np.any(selfie):
            facial_report['breakdown']['face_comparison']['breakdown']['face_match']['result'] = ''
            facial_report['breakdown']['face_comparison']['result'] = ''
            facial_report['breakdown']['image_integrity']['breakdown']['face_detected']['result'] = ''
            facial_report['breakdown']['image_integrity']['breakdown']['source_integrity']['result'] = ''

        if np.any(selfie) and facial_similarity >= face_match_threshold:
            facial_report['breakdown']['face_comparison']['breakdown']['face_match']['result'] = 'clear'
            facial_report['breakdown']['face_comparison']['result'] = 'clear'
            facial_report['breakdown']['image_integrity']['breakdown']['face_detected']['result'] = 'clear'
        elif facial_similarity != 0 and facial_similarity < face_match_threshold:
            facial_report['breakdown']['face_comparison']['breakdown']['face_match']['result']  = 'consider'
            facial_report['breakdown']['face_comparison']['result'] = 'consider'
            facial_report['breakdown']['image_integrity']['breakdown']['face_detected']['result'] = 'clear'
        else:
            facial_report['breakdown']['face_comparison']['breakdown']['face_match']['result']  = 'consider'
            facial_report['breakdown']['face_comparison']['result'] = 'consider'
            facial_report['breakdown']['image_integrity']['breakdown']['face_detected']['result'] = 'consider'

        front_photo_on_screen_result = data.get('front_photo_on_screen_result')

        if country == 'UAE':
            photo_on_screen_result = data.get('photo_on_screen_result')
            if front_photo_on_screen_result == 'consider' or photo_on_screen_result == 'consider':
                facial_report['breakdown']['image_integrity']['breakdown']['source_integrity']['result'] = 'consider'

        if country == 'SAU':
            facial_report['breakdown']['image_integrity']['breakdown']['source_integrity']['result'] = front_photo_on_screen_result

        facial_report['breakdown']['visual_authenticity']['breakdown']['spoofing_detection']['properties']['score'] = round(1-facial_similarity,2)

        facial_report['breakdown']['visual_authenticity']['breakdown']['liveness_detected']['result'] = 'clear'
        facial_report['breakdown']['visual_authenticity']['breakdown']['spoofing_detection']['result'] = 'clear'
        facial_report['breakdown']['visual_authenticity']['result'] = 'clear'

        if liveness_result == 'consider':
            facial_report['breakdown']['visual_authenticity']['breakdown']['liveness_detected']['result'] = 'consider'
            facial_report['breakdown']['visual_authenticity']['breakdown']['spoofing_detection']['result'] = 'consider'
            facial_report['breakdown']['visual_authenticity']['result'] = 'consider'
        elif liveness_result == 'clear' and facial_similarity < face_match_threshold:
            facial_report['breakdown']['visual_authenticity']['breakdown']['spoofing_detection']['result'] = 'consider'
            facial_report['breakdown']['visual_authenticity']['result'] = 'consider'
        elif liveness_result == None:
            facial_report['breakdown']['visual_authenticity']['breakdown']['liveness_detected']['result'] = ''
            facial_report['breakdown']['visual_authenticity']['breakdown']['spoofing_detection']['properties']['score'] = ''
            facial_report['breakdown']['visual_authenticity']['breakdown']['spoofing_detection']['result'] = ''

        # if np.any(selfie) and liveness_result:
        visual_authenticity_final_result = create_final_result(facial_report['breakdown']['visual_authenticity'])
        facial_report['breakdown']['visual_authenticity']['result'] = visual_authenticity_final_result

        # print(f"\n\n---------------- IMAGE INTEGRITY: {facial_report['breakdown']['image_integrity']}")
        # print(f"\n\n---------------- VISUAL AUTHENTICTY: {facial_report['breakdown']['visual_authenticity']}")
        # print(f"\n\n---------------- VISUAL AUTHENTICITY FINAL RESULT: {visual_authenticity_final_result}")

        image_integrity_final_result = create_final_result(facial_report['breakdown']['image_integrity'])
        facial_report['breakdown']['image_integrity']['result'] = image_integrity_final_result

        complete_final_result = create_final_result(facial_report['breakdown'])
        facial_report['result'] = complete_final_result
        facial_report['sub_result'] = complete_final_result

        return facial_report
    
    except Exception as e:
        print(f"\n--------------Error occurred while forming Facial Report: {str(e)}")
        raise Exception("Error occurred while creating facial_report")


"""Module to handle lazy loading of heavy dependencies"""

import psutil
import os

_deepface = None
_face_recognition = None
_pycountry = None
_translator = None

def log_memory_usage(operation):
    process = psutil.Process(os.getpid())
    memory_mb = process.memory_info().rss / 1024 / 1024
    print(f"Memory usage after {operation}: {memory_mb:.2f} MB")

def get_deepface():
    global _deepface
    if _deepface is None:
        log_memory_usage("before DeepFace import")
        from deepface import DeepFace
        _deepface = DeepFace
        log_memory_usage("after DeepFace import")
    return _deepface

def get_face_recognition():
    global _face_recognition
    if _face_recognition is None:
        import face_recognition
        _face_recognition = face_recognition
    return _face_recognition

def get_pycountry():
    global _pycountry
    if _pycountry is None:
        import pycountry
        _pycountry = pycountry
    return _pycountry

def get_translator():
    global _translator
    if _translator is None:
        from googletrans import Translator
        _translator = Translator()
    return _translator 


import pandas as pd
import re
from datetime import datetime
from hijri_converter import convert
from googletrans import Translator
import pycountry
import os

os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = "streamlit-connection-b1a38b694505 (1).json"

def get_country_code(name_in_foreign_language):
    translator = Translator()
    translated_country_name = translator.translate(name_in_foreign_language, src='auto', dest='en').text
    
    # Manual mapping for known challenges or discrepancies in translation
    manual_mapping = {
        'ميانمار': 'MMR',  # Myanmar
        'البحرين': 'BHR',  # Bahrain
        'اليمن الجنوبي': 'YEM',  # Yemen
        'فلسطين': 'PSE',  # Palestine
        'الفلبين': 'PHL',  # Philippines
        'أفغانستان': 'AFG',  # Afghanistan
        'سوريا': 'SYR',  # Syria
        'سري لنكا': 'LKA'  # Sri Lanka
    }
    
    if name_in_foreign_language in manual_mapping:
        return manual_mapping[name_in_foreign_language]
    
    try:
        country = pycountry.countries.get(name=translated_country_name)
        return country.alpha_3
    except AttributeError:
        return "Country Not Found"
    
pattern = r'(\d{4}/\d{1,2}/\d{1,2}|[۰-۹]{4}/[۰-۹]{1,2}/[۰-۹]{1,2})'
def eastern_arabic_to_english(eastern_numeral):
    arabic_to_english_map = {
        '٠': '0', '۰': '0',
        '١': '1', '۱': '1',
        '٢': '2', '۲': '2',
        '٣': '3', '۳': '3',
        '٤': '4', '۴': '4',
        '٥': '5', '۵': '5',
        '٦': '6', '۶': '6',
        '٧': '7', '۷': '7',
        '٨': '8', '۸': '8',
        '٩': '9', '۹': '9',
        '/': '/'
    }

    # If the character is an Eastern Arabic numeral, convert it to English; otherwise, keep it unchanged.
    english_numeral = ''.join([arabic_to_english_map[char] if char in arabic_to_english_map else char for char in eastern_numeral])
    
    return english_numeral

def distinguish_dates(date_list):
    today = datetime.now().date()
    
    # Calculate the difference between each date and today's date
    differences = [(abs((today - datetime.strptime(date, '%Y/%m/%d').date()).days), date) for date in date_list]
    
    # Sort by difference
    differences.sort(key=lambda x: x[0])
    
    # The date with the smallest difference is considered Gregorian, and the one with the largest difference is considered Hijri
    gregorian_date = differences[0][1]
    hijri_date = differences[-1][1]

    return hijri_date, gregorian_date

def hijri_to_gregorian(hijri_date):
    try:
        # Split the hijri date
        year, month, day = map(int, hijri_date.split('/'))

        # Convert the hijri date to Gregorian
        gregorian_date = convert.Hijri(year, month, day).to_gregorian()
    
    # Format the result as a string
        return f"{gregorian_date.year}/{gregorian_date.month:02}/{gregorian_date.day:02}"
    except:
        return hijri_date
    
def extract_dates(input_list):
    # Regex pattern to match YYYY/MM/DD, YYYY/MM/DD in Arabic numerals, 
    # and some other variations found in the list
    pattern = r"(\d{4}/\d{2}/\d{2}|[۰۱۲۳۴۵۶۷۸۹]{4}/[۰۱۲۳۴۵۶۷۸۹]{2}/[۰۱۲۳۴۵۶۷۸۹]{2})"
    
    extracted_dates = []
    for item in input_list:
        match = re.search(pattern, item)
        if match:
            extracted_dates.append(match.group(0))
        else:
            extracted_dates.append('')
    return extracted_dates


def detect_script(word):
    arabic_chars = range(0x0600, 0x06FF)  # Arabic Unicode Block
    english_chars = range(0x0041, 0x007A)  # English uppercase Unicode Block
    english_chars_lower = range(0x0061, 0x007A)  # English lowercase Unicode Block
    
    has_arabic = any(ord(char) in arabic_chars for char in word)
    has_english = any(ord(char) in english_chars or ord(char) in english_chars_lower for char in word)
    
    if has_arabic and has_english:
        return "Mixed"
    elif has_arabic:
        return "Arabic"
    elif has_english:
        return "English"
    else:
        return "Other"
    
def extract_english_strings(data):
    english_strings = []
    for string in data:
        if not re.search("[\u0600-\u06FF\d]", string):  # Filters out strings containing Arabic letters or digits
            english_strings.append(string)
    return english_strings

def extract_arabic_strings(data):
    arabic_strings = []
    for string in data:
        if re.search(r"[\u0600-\u06FF\d]", string):  # Filters out strings containing Arabic letters or digits
            arabic_strings.append(string)
    return arabic_strings

def clean_special_chars(data):
    cleaned_data = []
    for string in data:
        cleaned_string = re.sub(r'[^A-Za-z\s]', '', string)  # Retains only alphabets and spaces
        cleaned_data.append(cleaned_string.strip())  # .strip() removes any leading or trailing spaces
    return cleaned_data
    
def extract_id_details(result):
    # result = detect_text(uploaded_id)
    df = pd.DataFrame({'res':[result]})
    pattern = r'(\d{4}/\d{1,2}/\d{1,2}|[۰-۹]{4}/[۰-۹]{1,2}/[۰-۹]{1,2})'
    i = 0
    df['Extracted_data']=''
    try:
        nationality=[ele for ele in [ele for ele in df['res'].iloc[i] if 'الجنسية' in ele ][0].split('الجنسية') if ele!=''][0].strip()

        nationality=get_country_code(nationality)
    except:
        nationality=''
    try:
        ## employer
        employer_ar=[ele for ele in [ele for ele in df['res'].iloc[i] if 'صاحب العمل' in ele ]][0]
        employer=[ele for ele in employer_ar.split('صاحب العمل') if ele!=''][0].strip()

    except:
        employer=''
    try:
        ### issuing place
        issuing_place_ar=[ele for ele in df['res'].iloc[i] if 'مكان الإصدار' in ele][0]
        issuing_place=issuing_place_ar.split('مكان الإصدار')[-1].strip()
    except:
        issuing_place=''
    try:
        comon_pattern=[ele for ele in [ele for ele in df['res'].iloc[i] if (('الإصدار' in ele ) and('مكان' not in ele))][0].split('الإصدار') if ele!=''][0].strip()
        matches = re.findall(pattern, comon_pattern)

        matches=[eastern_arabic_to_english(ele) for ele in matches]

        issuing_date, dob=matches[0],matches[1]

        #issuing_date = hijri_to_gregorian(issuing_date)
        
    except:
        
        try: 
            dob=[ele for ele in [ele for ele in df['res'].iloc[i] if 'الميلاد' in ele ][0].split('الميلاد') if ele!=''][0].strip()
            issuing_date= [ele for ele in [ele for ele in df['res'].iloc[i] if( 'الإصدار' in ele) and ('مكان' not in ele ) ][0].split('الإصدار') if ele!=''][0].strip()
            #issuing_date=hijri_to_gregorian(issuing_date)
        except:
            try:
                dob=[ele for ele in [ele for ele in df['res'].iloc[i] if 'الميلاد' in ele ][0].split('الميلاد') if ele!=''][-1].strip()
                issuing_date=[ele for ele in [ele for ele in df['res'].iloc[i] if 'الميلاد' in ele ][0].split('الميلاد') if ele!=''][0].strip('الانتهاء').strip()
                #issuing_date=hijri_to_gregorian(issuing_date)
            except:
                issuing_date,dob='',''
    
    try:

        #issuing_date_ar,dob_ar=re.findall(pattern, comon_pattern)

        ### Id Number 
        id_number=[item for item in df['res'].iloc[i] if re.fullmatch(r'\d{10}', item)][0]
        id_number=eastern_arabic_to_english(id_number)
    
    except:
        
        try:
            id_number = [ele for ele in eastern_arabic_to_english([ele for ele in [ele for ele in df['res'].iloc[i] if 'الرقم' in ele ][0].split('الرقم') if ele!=''][0].strip()).split(' ') if len(ele)==10][0]
            id_number=eastern_arabic_to_english(id_number)
        except:
            id_number=''
    
    try:
        profession_Ar=[ele for ele in [ele for ele in df['res'].iloc[i] if 'المهنة' in ele ]][0]

        profession=[ele for ele in profession_Ar.split('المهنة') if ele!=''][-1]

    except:
        profession=''
    try:
        Name_Index=[extract_arabic_strings(df['res'].iloc[i]).index(ele) for ele in extract_arabic_strings(df['res'].iloc[i]) if 'وزارة' in ele][0]
        Name_1=extract_arabic_strings(df['res'].iloc[i])[Name_Index+1]
        Name_length=len(Name_1.split(' '))
        Name_en=max([ele for ele in clean_special_chars(extract_english_strings(df['res'].iloc[i])) if ele not in ['KINGDOM OF SAUDI ARABIA','MINISTRY OF INTERIOR']], key=lambda x: x.count(' '))
        Name_ar=[ele for ele in [Name_1,Name_en] if ele!=Name_en][0]  
        
    except:
        
        Name_en,Name_ar='',''

    df['Extracted_data'].iloc[i]=[nationality,employer,issuing_date, dob,id_number,profession,Name_en,Name_ar]
    
    cols = ['nationality', 'employer', 'issuing_date', 'dob', 'id_number', 'profession', 'Name_en', 'Name_ar']

    for index, col_name in enumerate(cols):
        df[col_name] = df['Extracted_data'].apply(lambda x: x[index])

    df['dob']=extract_dates(df['dob'].tolist())

    df['dob']=df['dob'].apply(lambda x: eastern_arabic_to_english(x))

    df['issuing_date']=df['issuing_date'].apply(lambda x: eastern_arabic_to_english(x))

    df['issuing_date']=df['issuing_date'].apply(lambda x: hijri_to_gregorian(x))

    dob = df['dob'].iloc[0]

    if dob:
        parsed_date = datetime.strptime(dob, "%Y/%m/%d")
        dob = parsed_date.strftime("%d/%m/%Y")

    # df['gender']= ''
    # df['expiry_data']= ' '

    # print(df)
    # TODO: gender, expiry_data
    return {'id_number': df['id_number'].iloc[0], 'nationality': df['nationality'].iloc[0], 'gender': '', 'dob': dob, 'expiry_date': '', 'name': df['Name_en'].iloc[0], 'occupation': df['profession'].iloc[0], 'employer': df['employer'].iloc[0]}




from __future__ import division
import torch
from PIL import Image, ImageOps, ImageEnhance
try:
    import accimage
except ImportError:
    accimage = None
import numpy as np
import numbers
import types
import collections
# import warnings


def _is_pil_image(img):
    if accimage is not None:
        return isinstance(img, (Image.Image, accimage.Image))
    else:
        return isinstance(img, Image.Image)


def _is_tensor_image(img):
    return torch.is_tensor(img) and img.ndimension() == 3


def _is_numpy_image(img):
    return isinstance(img, np.ndarray) and (img.ndim in {2, 3})


def to_tensor(pic):
    """Convert a ``PIL Image`` or ``numpy.ndarray`` to tensor.

    See ``ToTensor`` for more details.

    Args:
        pic (PIL Image or numpy.ndarray): Image to be converted to tensor.

    Returns:
        Tensor: Converted image.
    """
    if not(_is_pil_image(pic) or _is_numpy_image(pic)):
        raise TypeError('pic should be PIL Image or ndarray. Got {}'.format(type(pic)))

    if isinstance(pic, np.ndarray):
        # handle numpy array
        # IR image channel=1: modify by lzc --> 20190730
        if pic.ndim == 2:
            pic = pic.reshape((pic.shape[0], pic.shape[1], 1))

        img = torch.from_numpy(pic.transpose((2, 0, 1)))
        # backward compatibility
        # return img.float().div(255)  modify by zkx
        return img.float()
    if accimage is not None and isinstance(pic, accimage.Image):
        nppic = np.zeros([pic.channels, pic.height, pic.width], dtype=np.float32)
        pic.copyto(nppic)
        return torch.from_numpy(nppic)

    # handle PIL Image
    if pic.mode == 'I':
        img = torch.from_numpy(np.array(pic, np.int32, copy=False))
    elif pic.mode == 'I;16':
        img = torch.from_numpy(np.array(pic, np.int16, copy=False))
    else:
        img = torch.ByteTensor(torch.ByteStorage.from_buffer(pic.tobytes()))
    # PIL image mode: 1, L, P, I, F, RGB, YCbCr, RGBA, CMYK
    if pic.mode == 'YCbCr':
        nchannel = 3
    elif pic.mode == 'I;16':
        nchannel = 1
    else:
        nchannel = len(pic.mode)
    img = img.view(pic.size[1], pic.size[0], nchannel)
    # put it from HWC to CHW format
    # yikes, this transpose takes 80% of the loading time/CPU
    img = img.transpose(0, 1).transpose(0, 2).contiguous()
    if isinstance(img, torch.ByteTensor):
        # return img.float().div(255)  #modified by zkx
        return img.float()
    else:
        return img


def to_pil_image(pic, mode=None):
    """Convert a tensor or an ndarray to PIL Image.

    See :class:`~torchvision.transforms.ToPIlImage` for more details.

    Args:
        pic (Tensor or numpy.ndarray): Image to be converted to PIL Image.
        mode (`PIL.Image mode`_): color space and pixel depth of input data (optional).

    .. _PIL.Image mode: http://pillow.readthedocs.io/en/3.4.x/handbook/concepts.html#modes

    Returns:
        PIL Image: Image converted to PIL Image.
    """
    if not(_is_numpy_image(pic) or _is_tensor_image(pic)):
        raise TypeError('pic should be Tensor or ndarray. Got {}.'.format(type(pic)))

    npimg = pic
    if isinstance(pic, torch.FloatTensor):
        pic = pic.mul(255).byte()
    if torch.is_tensor(pic):
        npimg = np.transpose(pic.numpy(), (1, 2, 0))

    if not isinstance(npimg, np.ndarray):
        raise TypeError('Input pic must be a torch.Tensor or NumPy ndarray, ' +
                        'not {}'.format(type(npimg)))

    if npimg.shape[2] == 1:
        expected_mode = None
        npimg = npimg[:, :, 0]
        if npimg.dtype == np.uint8:
            expected_mode = 'L'
        if npimg.dtype == np.int16:
            expected_mode = 'I;16'
        if npimg.dtype == np.int32:
            expected_mode = 'I'
        elif npimg.dtype == np.float32:
            expected_mode = 'F'
        if mode is not None and mode != expected_mode:
            raise ValueError("Incorrect mode ({}) supplied for input type {}. Should be {}"
                             .format(mode, np.dtype, expected_mode))
        mode = expected_mode

    elif npimg.shape[2] == 4:
        permitted_4_channel_modes = ['RGBA', 'CMYK']
        if mode is not None and mode not in permitted_4_channel_modes:
            raise ValueError("Only modes {} are supported for 4D inputs".format(permitted_4_channel_modes))

        if mode is None and npimg.dtype == np.uint8:
            mode = 'RGBA'
    else:
        permitted_3_channel_modes = ['RGB', 'YCbCr', 'HSV']
        if mode is not None and mode not in permitted_3_channel_modes:
            raise ValueError("Only modes {} are supported for 3D inputs".format(permitted_3_channel_modes))
        if mode is None and npimg.dtype == np.uint8:
            mode = 'RGB'

    if mode is None:
        raise TypeError('Input type {} is not supported'.format(npimg.dtype))

    return Image.fromarray(npimg, mode=mode)


def normalize(tensor, mean, std):
    """Normalize a tensor image with mean and standard deviation.

    See ``Normalize`` for more details.

    Args:
        tensor (Tensor): Tensor image of size (C, H, W) to be normalized.
        mean (sequence): Sequence of means for each channel.
        std (sequence): Sequence of standard deviations for each channely.

    Returns:
        Tensor: Normalized Tensor image.
    """
    if not _is_tensor_image(tensor):
        raise TypeError('tensor is not a torch image.')

    for t, m, s in zip(tensor, mean, std):
        t.sub_(m).div_(s)
    return tensor


def resize(img, size, interpolation=Image.BILINEAR):
    """Resize the input PIL Image to the given size.

    Args:
        img (PIL Image): Image to be resized.
        size (sequence or int): Desired output size. If size is a sequence like
            (h, w), the output size will be matched to this. If size is an int,
            the smaller edge of the image will be matched to this number maintaing
            the aspect ratio. i.e, if height > width, then image will be rescaled to
            (size * height / width, size)
        interpolation (int, optional): Desired interpolation. Default is
            ``PIL.Image.BILINEAR``

    Returns:
        PIL Image: Resized image.
    """
    if not _is_pil_image(img):
        raise TypeError('img should be PIL Image. Got {}'.format(type(img)))
    if not (isinstance(size, int) or (isinstance(size, collections.Iterable) and len(size) == 2)):
        raise TypeError('Got inappropriate size arg: {}'.format(size))

    if isinstance(size, int):
        w, h = img.size
        if (w <= h and w == size) or (h <= w and h == size):
            return img
        if w < h:
            ow = size
            oh = int(size * h / w)
            return img.resize((ow, oh), interpolation)
        else:
            oh = size
            ow = int(size * w / h)
            return img.resize((ow, oh), interpolation)
    else:
        return img.resize(size[::-1], interpolation)


def scale(*args, **kwargs):
    # warnings.warn("The use of the transforms.Scale transform is deprecated, " +
    #               "please use transforms.Resize instead.")
    return resize(*args, **kwargs)


def pad(img, padding, fill=0):
    """Pad the given PIL Image on all sides with the given "pad" value.

    Args:
        img (PIL Image): Image to be padded.
        padding (int or tuple): Padding on each border. If a single int is provided this
            is used to pad all borders. If tuple of length 2 is provided this is the padding
            on left/right and top/bottom respectively. If a tuple of length 4 is provided
            this is the padding for the left, top, right and bottom borders
            respectively.
        fill: Pixel fill value. Default is 0. If a tuple of
            length 3, it is used to fill R, G, B channels respectively.

    Returns:
        PIL Image: Padded image.
    """
    if not _is_pil_image(img):
        raise TypeError('img should be PIL Image. Got {}'.format(type(img)))

    if not isinstance(padding, (numbers.Number, tuple)):
        raise TypeError('Got inappropriate padding arg')
    if not isinstance(fill, (numbers.Number, str, tuple)):
        raise TypeError('Got inappropriate fill arg')

    if isinstance(padding, collections.Sequence) and len(padding) not in [2, 4]:
        raise ValueError("Padding must be an int or a 2, or 4 element tuple, not a " +
                         "{} element tuple".format(len(padding)))

    return ImageOps.expand(img, border=padding, fill=fill)


def crop(img, i, j, h, w):
    """Crop the given PIL Image.

    Args:
        img (PIL Image): Image to be cropped.
        i: Upper pixel coordinate.
        j: Left pixel coordinate.
        h: Height of the cropped image.
        w: Width of the cropped image.

    Returns:
        PIL Image: Cropped image.
    """
    if not _is_pil_image(img):
        raise TypeError('img should be PIL Image. Got {}'.format(type(img)))

    return img.crop((j, i, j + w, i + h))


def center_crop(img, output_size):
    if isinstance(output_size, numbers.Number):
        output_size = (int(output_size), int(output_size))
    w, h = img.size
    th, tw = output_size
    i = int(round((h - th) / 2.))
    j = int(round((w - tw) / 2.))
    return crop(img, i, j, th, tw)


def resized_crop(img, i, j, h, w, size, interpolation=Image.BILINEAR):
    """Crop the given PIL Image and resize it to desired size.

    Notably used in RandomResizedCrop.

    Args:
        img (PIL Image): Image to be cropped.
        i: Upper pixel coordinate.
        j: Left pixel coordinate.
        h: Height of the cropped image.
        w: Width of the cropped image.
        size (sequence or int): Desired output size. Same semantics as ``scale``.
        interpolation (int, optional): Desired interpolation. Default is
            ``PIL.Image.BILINEAR``.
    Returns:
        PIL Image: Cropped image.
    """
    assert _is_pil_image(img), 'img should be PIL Image'
    img = crop(img, i, j, h, w)
    img = resize(img, size, interpolation)
    return img


def hflip(img):
    """Horizontally flip the given PIL Image.

    Args:
        img (PIL Image): Image to be flipped.

    Returns:
        PIL Image:  Horizontall flipped image.
    """
    if not _is_pil_image(img):
        raise TypeError('img should be PIL Image. Got {}'.format(type(img)))

    return img.transpose(Image.FLIP_LEFT_RIGHT)


def vflip(img):
    """Vertically flip the given PIL Image.

    Args:
        img (PIL Image): Image to be flipped.

    Returns:
        PIL Image:  Vertically flipped image.
    """
    if not _is_pil_image(img):
        raise TypeError('img should be PIL Image. Got {}'.format(type(img)))

    return img.transpose(Image.FLIP_TOP_BOTTOM)


def five_crop(img, size):
    """Crop the given PIL Image into four corners and the central crop.

    .. Note::
        This transform returns a tuple of images and there may be a
        mismatch in the number of inputs and targets your ``Dataset`` returns.

    Args:
       size (sequence or int): Desired output size of the crop. If size is an
           int instead of sequence like (h, w), a square crop (size, size) is
           made.
    Returns:
        tuple: tuple (tl, tr, bl, br, center) corresponding top left,
            top right, bottom left, bottom right and center crop.
    """
    if isinstance(size, numbers.Number):
        size = (int(size), int(size))
    else:
        assert len(size) == 2, "Please provide only two dimensions (h, w) for size."

    w, h = img.size
    crop_h, crop_w = size
    if crop_w > w or crop_h > h:
        raise ValueError("Requested crop size {} is bigger than input size {}".format(size,
                                                                                      (h, w)))
    tl = img.crop((0, 0, crop_w, crop_h))
    tr = img.crop((w - crop_w, 0, w, crop_h))
    bl = img.crop((0, h - crop_h, crop_w, h))
    br = img.crop((w - crop_w, h - crop_h, w, h))
    center = center_crop(img, (crop_h, crop_w))
    return (tl, tr, bl, br, center)


def ten_crop(img, size, vertical_flip=False):
    """Crop the given PIL Image into four corners and the central crop plus the
       flipped version of these (horizontal flipping is used by default).

    .. Note::
        This transform returns a tuple of images and there may be a
        mismatch in the number of inputs and targets your ``Dataset`` returns.

       Args:
           size (sequence or int): Desired output size of the crop. If size is an
               int instead of sequence like (h, w), a square crop (size, size) is
               made.
           vertical_flip (bool): Use vertical flipping instead of horizontal

        Returns:
            tuple: tuple (tl, tr, bl, br, center, tl_flip, tr_flip, bl_flip,
                br_flip, center_flip) corresponding top left, top right,
                bottom left, bottom right and center crop and same for the
                flipped image.
    """
    if isinstance(size, numbers.Number):
        size = (int(size), int(size))
    else:
        assert len(size) == 2, "Please provide only two dimensions (h, w) for size."

    first_five = five_crop(img, size)

    if vertical_flip:
        img = vflip(img)
    else:
        img = hflip(img)

    second_five = five_crop(img, size)
    return first_five + second_five


def adjust_brightness(img, brightness_factor):
    """Adjust brightness of an Image.

    Args:
        img (PIL Image): PIL Image to be adjusted.
        brightness_factor (float):  How much to adjust the brightness. Can be
            any non negative number. 0 gives a black image, 1 gives the
            original image while 2 increases the brightness by a factor of 2.

    Returns:
        PIL Image: Brightness adjusted image.
    """
    if not _is_pil_image(img):
        raise TypeError('img should be PIL Image. Got {}'.format(type(img)))

    enhancer = ImageEnhance.Brightness(img)
    img = enhancer.enhance(brightness_factor)
    return img


def adjust_contrast(img, contrast_factor):
    """Adjust contrast of an Image.

    Args:
        img (PIL Image): PIL Image to be adjusted.
        contrast_factor (float): How much to adjust the contrast. Can be any
            non negative number. 0 gives a solid gray image, 1 gives the
            original image while 2 increases the contrast by a factor of 2.

    Returns:
        PIL Image: Contrast adjusted image.
    """
    if not _is_pil_image(img):
        raise TypeError('img should be PIL Image. Got {}'.format(type(img)))

    enhancer = ImageEnhance.Contrast(img)
    img = enhancer.enhance(contrast_factor)
    return img


def adjust_saturation(img, saturation_factor):
    """Adjust color saturation of an image.

    Args:
        img (PIL Image): PIL Image to be adjusted.
        saturation_factor (float):  How much to adjust the saturation. 0 will
            give a black and white image, 1 will give the original image while
            2 will enhance the saturation by a factor of 2.

    Returns:
        PIL Image: Saturation adjusted image.
    """
    if not _is_pil_image(img):
        raise TypeError('img should be PIL Image. Got {}'.format(type(img)))

    enhancer = ImageEnhance.Color(img)
    img = enhancer.enhance(saturation_factor)
    return img


def adjust_hue(img, hue_factor):
    """Adjust hue of an image.

    The image hue is adjusted by converting the image to HSV and
    cyclically shifting the intensities in the hue channel (H).
    The image is then converted back to original image mode.

    `hue_factor` is the amount of shift in H channel and must be in the
    interval `[-0.5, 0.5]`.

    See https://en.wikipedia.org/wiki/Hue for more details on Hue.

    Args:
        img (PIL Image): PIL Image to be adjusted.
        hue_factor (float):  How much to shift the hue channel. Should be in
            [-0.5, 0.5]. 0.5 and -0.5 give complete reversal of hue channel in
            HSV space in positive and negative direction respectively.
            0 means no shift. Therefore, both -0.5 and 0.5 will give an image
            with complementary colors while 0 gives the original image.

    Returns:
        PIL Image: Hue adjusted image.
    """
    if not(-0.5 <= hue_factor <= 0.5):
        raise ValueError('hue_factor is not in [-0.5, 0.5].'.format(hue_factor))

    if not _is_pil_image(img):
        raise TypeError('img should be PIL Image. Got {}'.format(type(img)))

    input_mode = img.mode
    if input_mode in {'L', '1', 'I', 'F'}:
        return img

    h, s, v = img.convert('HSV').split()

    np_h = np.array(h, dtype=np.uint8)
    # uint8 addition take cares of rotation across boundaries
    with np.errstate(over='ignore'):
        np_h += np.uint8(hue_factor * 255)
    h = Image.fromarray(np_h, 'L')

    img = Image.merge('HSV', (h, s, v)).convert(input_mode)
    return img


def adjust_gamma(img, gamma, gain=1):
    """Perform gamma correction on an image.

    Also known as Power Law Transform. Intensities in RGB mode are adjusted
    based on the following equation:

        I_out = 255 * gain * ((I_in / 255) ** gamma)

    See https://en.wikipedia.org/wiki/Gamma_correction for more details.

    Args:
        img (PIL Image): PIL Image to be adjusted.
        gamma (float): Non negative real number. gamma larger than 1 make the
            shadows darker, while gamma smaller than 1 make dark regions
            lighter.
        gain (float): The constant multiplier.
    """
    if not _is_pil_image(img):
        raise TypeError('img should be PIL Image. Got {}'.format(type(img)))

    if gamma < 0:
        raise ValueError('Gamma should be a non-negative real number')

    input_mode = img.mode
    img = img.convert('RGB')

    np_img = np.array(img, dtype=np.float32)
    np_img = 255 * gain * ((np_img / 255) ** gamma)
    np_img = np.uint8(np.clip(np_img, 0, 255))

    img = Image.fromarray(np_img, 'RGB').convert(input_mode)
    return img


def rotate(img, angle, resample=False, expand=False, center=None):
    """Rotate the image by angle and then (optionally) translate it by (n_columns, n_rows)


    Args:
        img (PIL Image): PIL Image to be rotated.
        angle ({float, int}): In degrees degrees counter clockwise order.
        resample ({PIL.Image.NEAREST, PIL.Image.BILINEAR, PIL.Image.BICUBIC}, optional):
            An optional resampling filter.
            See http://pillow.readthedocs.io/en/3.4.x/handbook/concepts.html#filters
            If omitted, or if the image has mode "1" or "P", it is set to PIL.Image.NEAREST.
        expand (bool, optional): Optional expansion flag.
            If true, expands the output image to make it large enough to hold the entire rotated image.
            If false or omitted, make the output image the same size as the input image.
            Note that the expand flag assumes rotation around the center and no translation.
        center (2-tuple, optional): Optional center of rotation.
            Origin is the upper left corner.
            Default is the center of the image.
    """

    if not _is_pil_image(img):
        raise TypeError('img should be PIL Image. Got {}'.format(type(img)))

    return img.rotate(angle, resample, expand, center)


def to_grayscale(img, num_output_channels=1):
    """Convert image to grayscale version of image.

    Args:
        img (PIL Image): Image to be converted to grayscale.

    Returns:
        PIL Image:  Grayscale version of the image.
                    if num_output_channels == 1 : returned image is single channel
                    if num_output_channels == 3 : returned image is 3 channel with r == g == b
    """
    if not _is_pil_image(img):
        raise TypeError('img should be PIL Image. Got {}'.format(type(img)))

    if num_output_channels == 1:
        img = img.convert('L')
    elif num_output_channels == 3:
        img = img.convert('L')
        np_img = np.array(img, dtype=np.uint8)
        np_img = np.dstack([np_img, np_img, np_img])
        img = Image.fromarray(np_img, 'RGB')
    else:
        raise ValueError('num_output_channels should be either 1 or 3')

    return img



from __future__ import division
import math
import random
from PIL import Image
try:
    import accimage
except ImportError:
    accimage = None
import numpy as np
import numbers
import types

from idvpackage.spoof_resources import functional as F

__all__ = ["Compose", "ToTensor", "ToPILImage", "Normalize", "RandomHorizontalFlip",
           "Lambda", "RandomResizedCrop", "ColorJitter", "RandomRotation"]


class Compose(object):
    """Composes several transforms together.

    Args:
        transforms (list of ``Transform`` objects): list of transforms to compose.

    Example:
        >>> transforms.Compose([
        >>>     transforms.CenterCrop(10),
        >>>     transforms.ToTensor(),
        >>> ])
    """

    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, img):
        for t in self.transforms:
            img = t(img)
        return img


class ToTensor(object):

    """Convert a ``PIL Image`` or ``numpy.ndarray`` to tensor.

    Converts a PIL Image or numpy.ndarray (H x W x C) in the range
    [0, 255] to a torch.FloatTensor of shape (C x H x W) in the range [0.0, 1.0].
    """

    def __call__(self, pic):
        """
        Args:
            pic (PIL Image or numpy.ndarray): Image to be converted to tensor.

        Returns:
            Tensor: Converted image.
        """
        return F.to_tensor(pic)


class Lambda(object):
    """Apply a user-defined lambda as a transform.

    Args:
        lambd (function): Lambda/function to be used for transform.
    """

    def __init__(self, lambd):
        assert isinstance(lambd, types.LambdaType)
        self.lambd = lambd

    def __call__(self, img):
        return self.lambd(img)


class ToPILImage(object):
    """Convert a tensor or an ndarray to PIL Image.

    Converts a torch.*Tensor of shape C x H x W or a numpy ndarray of shape
    H x W x C to a PIL Image while preserving the value range.

    Args:
        mode (`PIL.Image mode`_): color space and pixel depth of input data (optional).
            If ``mode`` is ``None`` (default) there are some assumptions made about the input data:
            1. If the input has 3 channels, the ``mode`` is assumed to be ``RGB``.
            2. If the input has 4 channels, the ``mode`` is assumed to be ``RGBA``.
            3. If the input has 1 channel, the ``mode`` is determined by the data type (i,e,
            ``int``, ``float``, ``short``).

    .. _PIL.Image mode: http://pillow.readthedocs.io/en/3.4.x/handbook/concepts.html#modes
    """
    def __init__(self, mode=None):
        self.mode = mode

    def __call__(self, pic):
        """
        Args:
            pic (Tensor or numpy.ndarray): Image to be converted to PIL Image.

        Returns:
            PIL Image: Image converted to PIL Image.

        """
        return F.to_pil_image(pic, self.mode)


class Normalize(object):
    """Normalize an tensor image with mean and standard deviation.
    Given mean: ``(M1,...,Mn)`` and std: ``(S1,..,Sn)`` for ``n`` channels, this transform
    will normalize each channel of the input ``torch.*Tensor`` i.e.
    ``input[channel] = (input[channel] - mean[channel]) / std[channel]``

    Args:
        mean (sequence): Sequence of means for each channel.
        std (sequence): Sequence of standard deviations for each channel.
    """

    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, tensor):
        """
        Args:
            tensor (Tensor): Tensor image of size (C, H, W) to be normalized.

        Returns:
            Tensor: Normalized Tensor image.
        """
        return F.normalize(tensor, self.mean, self.std)


class RandomHorizontalFlip(object):
    """Horizontally flip the given PIL Image randomly with a probability of 0.5."""

    def __call__(self, img):
        """
        Args:
            img (PIL Image): Image to be flipped.

        Returns:
            PIL Image: Randomly flipped image.
        """
        if random.random() < 0.5:
            return F.hflip(img)
        return img


class RandomResizedCrop(object):
    """Crop the given PIL Image to random size and aspect ratio.

    A crop of random size (default: of 0.08 to 1.0) of the original size and a random
    aspect ratio (default: of 3/4 to 4/3) of the original aspect ratio is made. This crop
    is finally resized to given size.
    This is popularly used to train the Inception networks.

    Args:
        size: expected output size of each edge
        scale: range of size of the origin size cropped
        ratio: range of aspect ratio of the origin aspect ratio cropped
        interpolation: Default: PIL.Image.BILINEAR
    """

    def __init__(self, size, scale=(0.08, 1.0), ratio=(3. / 4., 4. / 3.), interpolation=Image.BILINEAR):
        if isinstance(size, tuple):
            self.size = size
        else:
            self.size = (size, size)
        self.interpolation = interpolation
        self.scale = scale
        self.ratio = ratio

    @staticmethod
    def get_params(img, scale, ratio):
        """Get parameters for ``crop`` for a random sized crop.

        Args:
            img (PIL Image): Image to be cropped.
            scale (tuple): range of size of the origin size cropped
            ratio (tuple): range of aspect ratio of the origin aspect ratio cropped

        Returns:
            tuple: params (i, j, h, w) to be passed to ``crop`` for a random
                sized crop.
        """
        for attempt in range(10):
            area = img.size[0] * img.size[1]
            target_area = random.uniform(*scale) * area
            aspect_ratio = random.uniform(*ratio)

            w = int(round(math.sqrt(target_area * aspect_ratio)))
            h = int(round(math.sqrt(target_area / aspect_ratio)))

            if random.random() < 0.5:
                w, h = h, w

            if w <= img.size[0] and h <= img.size[1]:
                i = random.randint(0, img.size[1] - h)
                j = random.randint(0, img.size[0] - w)
                return i, j, h, w

        # Fallback
        w = min(img.size[0], img.size[1])
        i = (img.size[1] - w) // 2
        j = (img.size[0] - w) // 2
        return i, j, w, w

    def __call__(self, img):
        """
        Args:
            img (PIL Image): Image to be flipped.

        Returns:
            PIL Image: Randomly cropped and resize image.
        """
        i, j, h, w = self.get_params(img, self.scale, self.ratio)
        return F.resized_crop(img, i, j, h, w, self.size, self.interpolation)


class ColorJitter(object):
    """Randomly change the brightness, contrast and saturation of an image.

    Args:
        brightness (float): How much to jitter brightness. brightness_factor
            is chosen uniformly from [max(0, 1 - brightness), 1 + brightness].
        contrast (float): How much to jitter contrast. contrast_factor
            is chosen uniformly from [max(0, 1 - contrast), 1 + contrast].
        saturation (float): How much to jitter saturation. saturation_factor
            is chosen uniformly from [max(0, 1 - saturation), 1 + saturation].
        hue(float): How much to jitter hue. hue_factor is chosen uniformly from
            [-hue, hue]. Should be >=0 and <= 0.5.
    """
    def __init__(self, brightness=0, contrast=0, saturation=0, hue=0):
        self.brightness = brightness
        self.contrast = contrast
        self.saturation = saturation
        self.hue = hue

    @staticmethod
    def get_params(brightness, contrast, saturation, hue):
        """Get a randomized transform to be applied on image.

        Arguments are same as that of __init__.

        Returns:
            Transform which randomly adjusts brightness, contrast and
            saturation in a random order.
        """
        transforms = []
        if brightness > 0:
            brightness_factor = np.random.uniform(max(0, 1 - brightness), 1 + brightness)
            transforms.append(Lambda(lambda img: F.adjust_brightness(img, brightness_factor)))

        if contrast > 0:
            contrast_factor = np.random.uniform(max(0, 1 - contrast), 1 + contrast)
            transforms.append(Lambda(lambda img: F.adjust_contrast(img, contrast_factor)))

        if saturation > 0:
            saturation_factor = np.random.uniform(max(0, 1 - saturation), 1 + saturation)
            transforms.append(Lambda(lambda img: F.adjust_saturation(img, saturation_factor)))

        if hue > 0:
            hue_factor = np.random.uniform(-hue, hue)
            transforms.append(Lambda(lambda img: F.adjust_hue(img, hue_factor)))

        np.random.shuffle(transforms)
        transform = Compose(transforms)

        return transform

    def __call__(self, img):
        """
        Args:
            img (PIL Image): Input image.

        Returns:
            PIL Image: Color jittered image.
        """
        transform = self.get_params(self.brightness, self.contrast,
                                    self.saturation, self.hue)
        return transform(img)


class RandomRotation(object):
    """Rotate the image by angle.

    Args:
        degrees (sequence or float or int): Range of degrees to select from.
            If degrees is a number instead of sequence like (min, max), the range of degrees
            will be (-degrees, +degrees).
        resample ({PIL.Image.NEAREST, PIL.Image.BILINEAR, PIL.Image.BICUBIC}, optional):
            An optional resampling filter.
            See http://pillow.readthedocs.io/en/3.4.x/handbook/concepts.html#filters
            If omitted, or if the image has mode "1" or "P", it is set to PIL.Image.NEAREST.
        expand (bool, optional): Optional expansion flag.
            If true, expands the output to make it large enough to hold the entire rotated image.
            If false or omitted, make the output image the same size as the input image.
            Note that the expand flag assumes rotation around the center and no translation.
        center (2-tuple, optional): Optional center of rotation.
            Origin is the upper left corner.
            Default is the center of the image.
    """

    def __init__(self, degrees, resample=False, expand=False, center=None):
        if isinstance(degrees, numbers.Number):
            if degrees < 0:
                raise ValueError("If degrees is a single number, it must be positive.")
            self.degrees = (-degrees, degrees)
        else:
            if len(degrees) != 2:
                raise ValueError("If degrees is a sequence, it must be of len 2.")
            self.degrees = degrees

        self.resample = resample
        self.expand = expand
        self.center = center

    @staticmethod
    def get_params(degrees):
        """Get parameters for ``rotate`` for a random rotation.

        Returns:
            sequence: params to be passed to ``rotate`` for random rotation.
        """
        angle = np.random.uniform(degrees[0], degrees[1])

        return angle

    def __call__(self, img):
        """
            img (PIL Image): Image to be rotated.

        Returns:
            PIL Image: Rotated image.
        """

        angle = self.get_params(self.degrees)

        return F.rotate(img, angle, self.resample, self.expand, self.center)





from datetime import datetime, timedelta
import re
from google.cloud import vision_v1
from googletrans import Translator
from deep_translator import GoogleTranslator
import io
from PIL import Image
import json
import openai
import time

translator = Translator()
import base64


def find_gender_from_back(text):
    gender = ''
    gender_pattern = r'(\d)([A-Za-z])(\d)'
    gender_match = re.search(gender_pattern, text)
    if gender_match:
        gender = gender_match.group(2)

    if not gender:
        gender_pattern = r'(\d)([MFmf])(\d)'
        gender_match = re.search(gender_pattern, text)
        if gender_match:
            gender = gender_match.group(2)

    return gender


def func_common_dates(extract_no_space):
    dob = ''
    expiry_date = ''
    try:
        matches = re.findall(r'\d{2}/\d{2}/\d{4}', extract_no_space)
        y1 = matches[0][-4:]
        y2 = matches[1][-4:]
        if int(y1) < int(y2):
            dob = matches[0]
            expiry_date = matches[1]
        else:
            dob = matches[1]
            expiry_date = matches[0]
    except:
        dob = ''
        expiry_date = ''

    return dob, expiry_date


def convert_dob(input_date):
    day = input_date[4:6]
    month = input_date[2:4]
    year = input_date[0:2]

    current_year = datetime.now().year
    current_century = current_year // 100
    current_year_last_two_digits = current_year % 100

    century = current_century
    # If the given year is greater than the last two digits of the current year, assume last century
    if int(year) > current_year_last_two_digits:
        century = current_century - 1

    final_date = f"{day}/{month}/{century}{year}"

    return final_date


def func_expiry_date(extract):
    extract_no_space = extract.replace(' ', '')
    dob, expiry_date = func_common_dates(extract_no_space)
    if expiry_date == '':
        match_doe = re.findall(r'\d{7}[A-Z]{2,3}', extract_no_space)
        for i in match_doe:

            raw_doe = i[0:6]
            print(raw_doe)
            expiry_date = raw_doe[4:6] + '/' + raw_doe[2:4] + '/20' + raw_doe[0:2]
            try:
                dt_obj = datetime.strptime(expiry_date, '%d/%m/%Y')
                break
            except:

                expiry_date = ''

    return expiry_date


def convert_expiry_date(input_date):
    day = input_date[4:6]
    month = input_date[2:4]
    year = input_date[0:2]

    current_year = datetime.now().year
    current_century = current_year // 100
    current_year_last_two_digits = current_year % 100
    century = current_century

    if int(year) <= current_year_last_two_digits:
        century = current_century
    else:
        century = current_century
    final_date = f"{day}/{month}/{century}{year}"

    return final_date


def func_dob(extract):
    extract_no_space = extract.replace(' ', '')
    dob, expiry_date = func_common_dates(extract_no_space)
    if dob == '':
        match_dob = re.findall(r'\d{7}(?:M|F)\d', extract_no_space)
        for i in match_dob:
            # print(i)
            raw_dob = i[0:6]
            # print(raw_dob)
            year = str(datetime.today().year)[2:4]
            temp = '19'
            if int(raw_dob[0:2]) > int(year):
                temp = '19'
            else:
                temp = '20'
            dob = raw_dob[4:6] + '/' + raw_dob[2:4] + '/' + temp + raw_dob[0:2]
            try:
                dt_obj = datetime.strptime(dob, '%d/%m/%Y')
                break
            except:
                # print(f'invalid date {dob}')
                dob = ''
        else:
            pattern = r"\b(\d{14}).*?\b"

            new_dob_match = re.search(pattern, extract_no_space)

            if new_dob_match:
                new_dob = new_dob_match.group(1)
                new_dob = new_dob[:7]
                dob = convert_dob(new_dob)

    return dob


def remove_special_characters_mrz2(string):
    # This pattern matches any character that is not a letter, digit, or space
    pattern = r'[^a-zA-Z0-9\s]'
    return re.sub(pattern, '', string)


def count_digits(element):
    digits = [char for char in element if char.isdigit()]
    return len(digits)


def sdn_back_id_extraction(back_id_data):
    mrz_pattern = r'(IDSDN.*\n*.*\n*.*\n*.*|IDSDN.*\n*.*\n*.*\n*.*|IDSDN.*\n*.*\n*.*\n*.*)'
    nationality_pattern = r'([A-Z]+)<<'

    mrz1, mrz2, mrz3 = '', '', ''

    try:
        mrz = re.findall(mrz_pattern, back_id_data.replace(" ", "").strip(), re.MULTILINE)
        mrz_str = mrz[0].replace(" ", "")
    except:
        mrz_str = ''

    if mrz_str:
        mrz_list = mrz_str.replace(" ", "").split("\n")
        try:
            mrz1 = mrz_list[0]
            if len(mrz_list) == 3:
                mrz1, mrz2, mrz3 = mrz_list[0], mrz_list[1], mrz_list[2]
        except:
            mrz1 = ''

        try:
            mrz2 = [ele for ele in [ele for ele in mrz_list if ele not in [mrz1, mrz3]] if
                    remove_special_characters_mrz2(ele) != '']
            if len(mrz2) > 1:
                mrz2 = max(mrz2, key=count_digits) + [ele for ele in mrz2 if ele != max(mrz2, key=count_digits)][0]

                pattern = r'\d{7}[MF]\d{7}[\S]{3}<+?\d'
                mrz2_temp = re.search(pattern, mrz2.replace(">", ""))
                if mrz2_temp:
                    mrz2 = mrz2_temp.group(0)

                mrz2 = mrz2.split('<')[0] + '<<<<<<<<<<' + mrz2.split('<')[-1]

                # mrz2=mrz2[0].split('<')[0]+'<<<<<<<<<<'+mrz2[-1][-1]
            else:
                mrz2 = mrz2[0].split('<')[0] + '<<<<<<<<<<' + mrz2[0][-1]
        except:
            mrz2 = ''

    ## condition to replace O with 0
    try:
        pattern = r'(IDSDN[A-Z]{1,2})O(?=[0-9])'
        replacement = lambda m: m.group(1) + '0'
        mrz1 = re.sub(pattern, replacement, mrz1)
    except:
        pass

    ## condition to replace '>' with 7
    if mrz2 and mrz2.endswith('>'):
        mrz2 = mrz2.split('<')[0] + '<<<<<<<<<<' + '7'

    if not mrz3 or (mrz3.startswith('>') or mrz3.startswith('<')):
        pattern = r'^[A-Za-z]+<+[A-Za-z]+.*$'
        matches = re.findall(pattern, mrz_str, re.MULTILINE)
        try:
            mrz3 = list(filter(None, matches))[0]
        except:
            try:
                matches = re.findall(pattern, back_id_data, re.MULTILINE)
                mrz3 = list(filter(None, matches))[0]
            except:
                mrz3 = ''

    ## condition to add filler to mrz3, making it total length of 30 chars
    if len(mrz3) < 30:
        mrz3 = mrz3.ljust(30, '<')

    try:
        dob = func_dob(mrz2)
    except:
        dob = ''

    if not dob:
        matches = re.findall(r'\d{4}/\d{2}/\d{2}', back_id_data)
        sorted_dates = sorted(matches)
        dob = sorted_dates[0]

    expiry = func_expiry_date(mrz_str)
    if not expiry:
        matches = re.findall(r'\d{4}/\d{2}/\d{2}', back_id_data)
        sorted_dates = sorted(matches)
        expiry = sorted_dates[-1]

    # issue date
    issue_date = ''  # Initialize with default value
    try:
        matches = re.findall(r'\d{4}/\d{2}/\d{2}', back_id_data)
        sorted_dates = sorted(matches)
        if len(sorted_dates) > 2:
            issue_date = sorted_dates[1]
    except:
        pass

    try:
        nationality = mrz2.split('<')[0][-3:]
    except:
        nationality = ''

    if mrz3:
        full_name_mrz = mrz3.replace('<', ' ').replace('>', ' ').strip()

    else:
        full_name_mrz = ''

    try:
        pattern = r'(?<=Name: )\w+(?: \w+)*|(?<=Name )\w+(?: \w+)*'

        match = re.search(pattern, back_id_data, re.IGNORECASE)
        name = match.group(0) or match.group(1)
    except:
        try:
            pattern = r'(?<=NAME):*[ \n]*([A-Z ]+)'

            match = re.search(pattern, back_id_data, re.IGNORECASE)
            if match:
                name = match.group(1).strip().replace(':', '')
            else:
                name = ''
        except:
            name = ''

    if full_name_mrz and not name:
        name = ' '.join(full_name_mrz.split(' ')[1:]) + ' ' + full_name_mrz.split(' ')[0] if full_name_mrz else ''
        name = name.strip()

    if name:
        first_name = name.split(' ')[0]
        last_name = name.split(' ')[-1]
        middle_name = ' '.join(name.split(' ')[1:-1])
    else:
        first_name, last_name, middle_name = '', '', ''

    if 'issue_date' not in locals():
        issue_date = ''

    try:
        dob_pattern = r'(\d+)[MF]'
        dob_match = re.search(dob_pattern, mrz2)
        dob_mrz = convert_dob(dob_match.group(1)) if dob_match else ''

        doe_pattern = r'[MF](\d+)'
        doe_match = re.search(doe_pattern, mrz2)
        expiry_date_mrz = convert_expiry_date(doe_match.group(1)) if doe_match else ''
    except:
        dob_mrz, expiry_date_mrz = '', ''

    gender = ''
    try:
        gender = find_gender_from_back(mrz2)
    except:
        gender = find_gender_from_back(back_id_data)

    mrz_str = f"{mrz1}\n{mrz2}\n{mrz3}"

    try:
        if expiry and not expiry_date_mrz:
            expiry_date_mrz = expiry

        if dob and not dob_mrz:
            dob_mrz = dob
    except:
        pass

    if issue_date == '':
        print(f"Calculating issue date....")
        from dateutil.relativedelta import relativedelta
        try:
            exp = datetime.strptime(expiry_date_mrz, "%d/%m/%Y")
        except:
            exp = datetime.strptime(expiry_date_mrz, "%d-%m-%Y")

        issue_date = exp - relativedelta(years=5) + timedelta(days=1)
        issue_date = issue_date.strftime("%d/%m/%Y")

    back_data_dict = {
        "mrz": [mrz_str],
        "mrz1": mrz1.replace('*', '<'),
        "mrz2": mrz2,
        "mrz3": mrz3,
        # "dob_generic": dob,
        # "full_name_mrz": full_name_mrz,
        "full_name_generic": name,
        "first_name": first_name,
        "middle_name": middle_name,
        "last_name": last_name,
        "issuing_country": "SDN",
        # "expiry_date_generic": expiry,
        "nationality": nationality,
        "dob_back": dob_mrz,
        "issue_date": issue_date,
        "expiry_date": expiry_date_mrz,
        "gender": gender
    }
    if "gender" in back_data_dict:
        gender = back_data_dict["gender"].strip().upper()
        if gender == "F":
            back_data_dict["gender"] = "FEMALE"
        elif gender == "M":
            back_data_dict["gender"] = "MALE"

    return back_data_dict


def crop_second_part(img):
    width, height = img.size
    half_width = width // 2
    second_part = img.crop((half_width, 0, width, height))
    return second_part


def extract_text_from_image_data(client, image):
    """Detects text in the file."""

    with io.BytesIO() as output:
        image.save(output, format="PNG")
        content = output.getvalue()

    image = vision_v1.types.Image(content=content)

    response = client.text_detection(image=image)
    texts = response.text_annotations

    return texts[0].description


def detect_id_card(client, image_data, id_text, part=None):
    if id_text:
        vertices = id_text[0].bounding_poly.vertices
        left = vertices[0].x
        top = vertices[0].y
        right = vertices[2].x
        bottom = vertices[2].y

        padding = 30
        left -= padding
        top -= padding
        right += padding
        bottom += padding

        # img = image_data

        with Image.open(io.BytesIO(image_data)) as img:
            id_card = img.crop((max(0, left), max(0, top), right, bottom))
            width, height = id_card.size
            if width < height:
                id_card = id_card.rotate(90, expand=True)

            part_text = id_text[0].description
            part_img = crop_second_part(id_card)
            part_text = extract_text_from_image_data(client, part_img)

            return part_text


def extract_occupation(text):
    match = re.search(r'المهنة\s*([^\n]+)', text)
    if match:
        return match.group(1).strip().replace(":", "")
    else:
        match = re.search(r'المهن[ةــ]*\s*\n\s*([^\n]+)', text)
        if match:
            return match.group(1).replace(":", "").replace("ــة", "").replace("ة.", "").replace("ة", "").replace(
                "العقــ", "").replace("ـ", "").strip()
        else:
            return None


def extract_occupation_from_text(part_text):
    lines = part_text.split('\n')

    arabic_number_pattern = re.compile(r'[\u0660-\u0669]+')
    final_occupation = ''

    for i in range(len(lines) - 1, 0, -1):
        if arabic_number_pattern.search(lines[i]):
            occupation = lines[i - 1].strip()
            occupation = re.sub(r'\d+', '', occupation)
            occupation = re.sub(r'[A-Za-z]+', '', occupation)
            if occupation and occupation in ['العنــ', 'الغد', 'المهنــ', 'العنوان', 'العلب', 'العنب', 'ـوان', 'العيد',
                                             'العز', 'العن', 'العد']:
                search_key = lines.index(occupation)
                final_occupation = lines[search_key - 1]
                break
            else:
                final_occupation = occupation

    return final_occupation.replace(":", "").replace("المهنة", "").replace("ــة", "").replace("ة.", "").replace("ة",
                                                                                                                "").replace(
        "العقــ", "").replace("ـ", "").strip() if final_occupation else ""


def extract_occupation_v2(client, image_data, texts):
    part_text = detect_id_card(client, image_data, texts)
    occupation_res = extract_occupation_from_text(part_text)
    if not occupation_res:
        occupation_res = extract_occupation_from_text(texts[0].description)

    if occupation_res in ['الرقم الوطني'] or not occupation_res:
        occupation_res = extract_occupation(part_text)
        if not occupation_res:
            occupation_res = extract_occupation(texts[0].description)

    return occupation_res


def extract_place_of_birth(text):
    match = re.search(r'مكان الميلاد\s*([^\n]+)|مكان الميادد\s*([^\n]+)', text)
    if match:
        return match.group(1).strip().replace(":", "") if match.group(1) is not None else match.group(
            2).strip().replace(":", "")
    return None


def extract_dob(text):
    dob = ''
    extract_no_space = text.replace(" ", "")
    try:
        matches = re.findall(r'\d{4}/\d{2}/\d{2}', extract_no_space)
        dob = matches[0]
    except:
        dob = ''
    return dob


def extract_name_from_front(text, dob):
    name = ''
    lines = text.split("\n")
    for line in lines:
        if dob in line:
            search_key = lines.index(line)
            part_name = lines[search_key - 1]
            if part_name:
                part_name = re.sub(r'\d+', '', part_name)
                part_name = re.sub(r'[A-Za-z]+', '', part_name)
                name = part_name.replace("الإســــــــــــم", "").replace("الاسم", "").replace("الإسم", "").replace(
                    "ـم", "").replace(":", "").strip()
                if name in ['تاريخ الميلاد', 'الإســ', 'الإسـ', 'الإس', 'الإصـ', 'تاریخ', 'الرقم الوطني'] or len(
                        name.split(' ')) <= 2:
                    name = lines[search_key - 2]
                    if name:
                        name = name.replace("الإســــــــــــم", "").replace("الاسم", "").replace("الإسم", "").replace(
                            "ـم", "").replace(":", "").strip()
                        name = re.sub(r'\d+', '', name)
                        name = re.sub(r'[A-Za-z]+', '', name)
                break
    return name.strip()


def sdn_front_id_extraction(client, ar_front_id_data, image_data, texts, compressed_image):
    front_id_data = translator.translate(ar_front_id_data, src='ar', dest='en').text
    # print(f'\n\nTranslated front ID data: {front_id_data}\n')

    id_number_pattern = r"\b\d{11}\b"

    id_number_match = re.search(id_number_pattern, front_id_data.replace(" ", ""), re.IGNORECASE)
    if id_number_match:
        id_number = id_number_match.group(0)
    else:
        try:
            id_number_match = re.search(id_number_pattern, ar_front_id_data.replace(" ", ""), re.IGNORECASE)
            id_number = id_number_match.group(0)
        except:
            id_number = ''

    try:
        occupation = extract_occupation_v2(client, image_data, texts)
        if occupation:
            try:
                occupation_en = GoogleTranslator(dest='en').translate(occupation)
            except:
                occupation_en = ''
        else:
            occupation, occupation_en = '', ''
    except:
        occupation, occupation_en = '', ''

    try:
        place_of_birth = extract_place_of_birth(ar_front_id_data)
        if place_of_birth:
            try:
                place_of_birth_en = GoogleTranslator(dest='en').translate(place_of_birth)
            except:
                place_of_birth_en = ''
        else:
            place_of_birth, place_of_birth_en = '', ''
    except:
        place_of_birth, place_of_birth_en = '', ''

    try:
        dob = extract_dob(ar_front_id_data)
    except:
        dob = ''

    try:
        full_name = extract_name_from_front(ar_front_id_data, dob)
        if full_name:
            if 'مكان الميلاد' in full_name:
                lines = ar_front_id_data.split('\n')
                search_pos = lines.index(full_name)
                full_name = lines[search_pos - 1]
        else:
            full_name
    except:
        full_name

    front_data_dict = {
        'id_number': id_number,
        'occupation_ar': occupation,
        'occupation_en': occupation_en,
        'occupation': occupation_en,
        'place_of_birth': place_of_birth,
        'place_of_birth_en': place_of_birth_en,
        'dob': dob,
        'name_ar': full_name
    }

    empty_string_keys = [key for key, value in front_data_dict.items() if value == '']
    if empty_string_keys:
        prompt = """
From the provided text: " %s ", extract and structure the following fields as a dictionary:

- 'id_number': The ID number (e.g., national ID, passport number, etc.)
- 'occupation_ar': The occupation in Arabic
- 'occupation_en': The occupation in English
- 'place_of_birth': The place of birth in Arabic
- 'place_of_birth_en': The place of birth in English
- 'dob': The date of birth (in the format YYYY-MM-DD or any standard date format provided)
- 'name_ar': The full name in Arabic

The response should STRICTLY follow this format:
{
  "id_number": "<value>",
  "occupation_ar": "<value>",
  "occupation_en": "<value>",
  "place_of_birth": "<value>",
  "place_of_birth_en": "<value>",
  "dob": "<value>",
  "name_ar": "<value>"
}
Ensure that all values are accurately extracted and formatted. If a value is missing, return `null` for that field.

Example:
{
  "id_number": "12345678901",
  "occupation_ar": "مهندس",
  "occupation_en": "Engineer",
  "place_of_birth": "الرياض",
  "place_of_birth_en": "Riyadh",
  "dob": "1990-05-15",
  "name_ar": "محمد بن أحمد"
}
""" % front_id_data

        start = time.time()
        front_data_dict = get_openai_response_with_retries(prompt=prompt, compressed_image=compressed_image)
        end = time.time() - start
        print(f'Openai api call took an additional {end}s')
        front_data_dict['occupation'] = front_data_dict['occupation_en']

    return front_data_dict


def get_openai_response_with_retries(max_retries=3, prompt='', delay_seconds: float = 2, compressed_image=''):
    img_bytes = compressed_image.getvalue()

    # Encode the bytes to base64
    img_base64_bytes = base64.b64encode(img_bytes).decode("utf-8")
    for attempt in range(max_retries):
        try:
            response = openai.ChatCompletion.create(
                model='gpt-4.1-nano',
                temperature=0.4,
                max_tokens=2000,
                messages=[
                    {
                        "role": "user",
                        "content": [
                            {"type": "text", "text": prompt},
                            {
                                "type": "image_url",
                                "image_url": {
                                    "url": f"data:image/jpeg;base64,{img_base64_bytes}",
                                },
                            },
                        ],
                    }
                ],
            )

            result = response.choices[0].message.content

            try:
                return json.loads(result)
            except json.JSONDecodeError:
                try:
                    json_match = re.search(r'```(json|python|plaintext)?\s*(.*?)\s*```|\s*({.*?})', result,
                                           re.DOTALL)
                    if json_match:
                        json_str = json_match.group(2) or json_match.group(3)
                        try:
                            return json.loads(json_str)
                        except:
                            return eval(json_str.replace("'", '"'))
                except Exception as e:
                    return {'error': "GPT's response incorrectly formatted.", 'error_details': e}


        except Exception as e:

            print(f"Error during API request (attempt {attempt + 1} of {max_retries}): {str(e)}")

            if attempt < max_retries - 1:

                time.sleep(delay_seconds)

            else:

                raise Exception(f"Max retries exceeded. Last error: {str(e)}")




"""
Create patch from original input image by using bbox coordinate
"""

import cv2
import numpy as np


class CropImage:
    @staticmethod
    def _get_new_box(src_w, src_h, bbox, scale):
        x = bbox[0]
        y = bbox[1]
        box_w = bbox[2]
        box_h = bbox[3]

        scale = min((src_h-1)/box_h, min((src_w-1)/box_w, scale))

        new_width = box_w * scale
        new_height = box_h * scale
        center_x, center_y = box_w/2+x, box_h/2+y

        left_top_x = center_x-new_width/2
        left_top_y = center_y-new_height/2
        right_bottom_x = center_x+new_width/2
        right_bottom_y = center_y+new_height/2

        if left_top_x < 0:
            right_bottom_x -= left_top_x
            left_top_x = 0

        if left_top_y < 0:
            right_bottom_y -= left_top_y
            left_top_y = 0

        if right_bottom_x > src_w-1:
            left_top_x -= right_bottom_x-src_w+1
            right_bottom_x = src_w-1

        if right_bottom_y > src_h-1:
            left_top_y -= right_bottom_y-src_h+1
            right_bottom_y = src_h-1

        return int(left_top_x), int(left_top_y),\
               int(right_bottom_x), int(right_bottom_y)

    def crop(self, org_img, bbox, scale, out_w, out_h, crop=True):

        if not crop:
            dst_img = cv2.resize(org_img, (out_w, out_h))
        else:
            src_h, src_w, _ = np.shape(org_img)
            left_top_x, left_top_y, \
                right_bottom_x, right_bottom_y = self._get_new_box(src_w, src_h, bbox, scale)

            img = org_img[left_top_y: right_bottom_y+1,
                          left_top_x: right_bottom_x+1]
            dst_img = cv2.resize(img, (out_w, out_h))
        return dst_img
